@@ -1,6 +1,7 @@
 // cmdline_do_action.cc
 //
 //  Copyright (C) 2004, 2010 Daniel Burrows
+//  Copyright (C) 2014-2016 Manuel A. Fernandez Montecelo
 //
 // This program is free software; you can redistribute it and/or
 // modify it under the terms of the GNU General Public License as
@@ -14,8 +15,8 @@
 //
 // You should have received a copy of the GNU General Public License
 // along with this program; see the file COPYING.  If not, write to
-// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-// Boston, MA 02111-1307, USA.
+// the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
+// Boston, MA 02110-1301, USA.
 
 
 // Local includes:
@@ -53,6 +54,12 @@ using aptitude::cmdline::create_terminal;
 using aptitude::cmdline::make_text_progress;
 using aptitude::cmdline::terminal_io;
 
+
+void print_clean_msg()
+{
+  printf("%s\n", _("Deleting downloaded files"));
+}
+
 namespace
 {
   void run_dpkg_directly(sigc::slot1<pkgPackageManager::OrderResult, int> f,
@@ -138,21 +145,22 @@ int cmdline_do_action(int argc, char *argv[],
     default_action=cmdline_remove;
   else if(!strcasecmp(argv[0], "purge"))
     default_action=cmdline_purge;
-  else if(!strcasecmp(argv[0], "hold"))
-    default_action=cmdline_hold;
-  else if(!strcasecmp(argv[0], "keep") || !strcasecmp(argv[0], "keep-all"))
-    default_action=cmdline_keep;
-  else if(!strcasecmp(argv[0], "unhold"))
-    default_action=cmdline_unhold;
-  else if(!strcasecmp(argv[0], "markauto"))
-    default_action=cmdline_markauto;
-  else if(!strcasecmp(argv[0], "unmarkauto"))
-    default_action=cmdline_unmarkauto;
-  else if(!strcasecmp(argv[0], "forbid-version"))
-    default_action=cmdline_forbid_version;
   else if(!strcasecmp(argv[0], "build-depends") ||
 	  !strcasecmp(argv[0], "build-dep"))
     default_action = cmdline_build_depends;
+  else if (!strcasecmp(argv[0], "hold") ||
+	   !strcasecmp(argv[0], "keep") ||
+	   !strcasecmp(argv[0], "keep-all") ||
+	   !strcasecmp(argv[0], "unhold") ||
+	   !strcasecmp(argv[0], "markauto") ||
+	   !strcasecmp(argv[0], "unmarkauto") ||
+	   !strcasecmp(argv[0], "forbid-version"))
+    {
+      // should happen elsewhere
+      _error->Error(_("Internal error, operation %s should be handled elsewhere"), argv[0]);
+      _error->DumpErrors();
+      return -1;
+    }
   else
     {
       // Should never happen.
@@ -173,9 +181,11 @@ int cmdline_do_action(int argc, char *argv[],
   //
   // This way, "aptitude install" will just perform any pending
   // installations.
+  bool operation_needs_lock = true;
   apt_init(progress.get(),
-           (argc==1 && default_action==cmdline_install &&
-            upgrade_mode == no_upgrade), status_fname);
+           (argc==1 && default_action==cmdline_install && upgrade_mode == no_upgrade),
+	   operation_needs_lock,
+	   status_fname);
 
   if(_error->PendingError())
     {
@@ -199,7 +209,7 @@ int cmdline_do_action(int argc, char *argv[],
   ReadPinFile(policy);
   ReadPinDir(policy);
 
-  pkgset to_upgrade, to_install, to_hold, to_remove, to_purge;
+  pkgset to_install, to_hold, to_remove, to_purge;
 
   if(upgrade_mode == full_upgrade || upgrade_mode == safe_upgrade)
     {
@@ -210,16 +220,30 @@ int cmdline_do_action(int argc, char *argv[],
       bool use_autoinst = (resolver_mode != resolver_mode_safe);
       (*apt_cache_file)->mark_all_upgradable(use_autoinst, ignore_removed, NULL);
     }
-  /*else if(argc==1 && default_action==cmdline_install)
+
+  // same implementation in CmdLine full-upgrade and UI MarkUpgradable
+  //
+  // install Essential/Required packages -- #555896 and #757028
+  if (upgrade_mode == full_upgrade)
     {
-      // FIXME: Build to_install to avoid a big printout
-      for(pkgCache::PkgIterator i=(*apt_cache_file)->PkgBegin(); !i.end(); ++i)
+      for (pkgCache::GrpIterator grp = (*apt_cache_file)->GrpBegin(); !grp.end(); ++grp)
 	{
+	  for (pkgCache::PkgIterator pkg = grp.PackageList(); !pkg.end(); pkg = grp.NextPkg(pkg))
+	    {
+	      bool is_essential = (pkg->Flags & pkgCache::Flag::Essential) == pkgCache::Flag::Essential;
+	      if (!is_essential || is_installed(pkg))
+		continue;
 
+	      pkgCache::PkgIterator preferred_pkg = grp.FindPreferredPkg();
+	      if (is_installed(preferred_pkg) || (*apt_cache_file)[preferred_pkg].Install())
+		continue;
+
+	      (*apt_cache_file)->mark_install(preferred_pkg, false, false, nullptr);
+	    }
 	}
-	}*/
+    }
 
-  // TODO: look for filenames and call dpkg directly if that's the case.
+  bool apply_ok = true;
 
   {
     aptitudeDepCache::action_group group(*apt_cache_file, NULL);
@@ -325,15 +349,21 @@ int cmdline_do_action(int argc, char *argv[],
 	  for(std::vector<action_pair>::const_iterator it = actions.begin();
 	      it != actions.end(); ++it)
 	    {
-	      cmdline_applyaction(it->second, seen_virtual_packages, it->first,
-				  to_install, to_hold, to_remove, to_purge,
-				  verbose, policy, arch_only, pass > 0,
-                                  term);
+	      apply_ok = apply_ok && cmdline_applyaction(it->second, seen_virtual_packages, it->first,
+							 to_install, to_hold, to_remove, to_purge,
+							 verbose, policy, arch_only, pass > 0,
+							 term);
 	    }
 	}
     }
   }
 
+  if (!apply_ok)
+    {
+      fprintf(stderr, _("Unable to apply some actions, aborting\n"));
+      return -1;
+    }
+
   if(resolver_mode == resolver_mode_safe)
     {
       if(!aptitude::cmdline::safe_resolve_deps(verbose,
@@ -359,13 +389,13 @@ int cmdline_do_action(int argc, char *argv[],
 			    showsize, showwhy,
 			    always_prompt, verbose, assume_yes,
 			    !fix_broken,
-			    policy, arch_only,
+			    policy, arch_only, download_only,
                             term);
   else if(queue_only)
     {
       aptitude::cmdline::apply_user_tags(user_tags);
 
-      if(!(*apt_cache_file)->save_selection_list(*progress))
+      if(!(*apt_cache_file)->save_selection_list(progress.get()))
 	{
 	  _error->DumpErrors();
 	  return -1;
@@ -380,7 +410,7 @@ int cmdline_do_action(int argc, char *argv[],
 			    showvers, showdeps, showsize, showwhy,
 			    always_prompt, verbose, assume_yes,
 			    !fix_broken,
-			    policy, arch_only, term))
+			    policy, arch_only, download_only, simulate, term))
 	{
 	  printf(_("Abort.\n"));
 	  return 1;
@@ -391,6 +421,8 @@ int cmdline_do_action(int argc, char *argv[],
       download_install_manager m(download_only,
 				 sigc::ptr_fun(&run_dpkg_directly));
 
+      m.pre_clean_after_install_hook.connect(sigc::ptr_fun(print_clean_msg));
+
       int rval =
 	(cmdline_do_download(&m, verbose, term, term, term, term)
          == download_manager::success ? 0 : -1);
