@@ -1,6 +1,7 @@
 // cmdline_prompt.cc
 //
 // Copyright (C) 2010-2011 Daniel Burrows
+// Copyright (C) 2014-2016 Manuel A. Fernandez Montecelo
 //
 // This program is free software; you can redistribute it and/or
 // modify it under the terms of the GNU General Public License as
@@ -14,8 +15,8 @@
 //
 // You should have received a copy of the GNU General Public License
 // along with this program; see the file COPYING.  If not, write to
-// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-// Boston, MA 02111-1307, USA.
+// the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
+// Boston, MA 02110-1301, USA.
 //
 //   Handles the preview and prompt that's displayed from the command-line.
 
@@ -45,9 +46,7 @@
 
 // System includes:
 #include <apt-pkg/algorithms.h>
-#include <apt-pkg/dpkgpm.h>
 #include <apt-pkg/error.h>
-#include <apt-pkg/sourcelist.h>
 #include <apt-pkg/strutl.h>
 
 #include <cwidget/fragment.h>
@@ -64,35 +63,6 @@ using aptitude::why::make_cmdline_why_callbacks;
 using aptitude::why::why_callbacks;
 
 
-struct fetchinfo
-{
-  unsigned long long FetchBytes, FetchPBytes, DebBytes;
-
-  fetchinfo()
-    : FetchBytes(0), FetchPBytes(0), DebBytes(0)
-  {
-  }
-};
-
-static bool get_fetchinfo(fetchinfo &f)
-{
-  download_signal_log m;
-  pkgAcquire fetcher;
-  fetcher.Setup(&m);
-  pkgSourceList l;
-  if(!l.ReadMainList())
-    return _error->Error(_("Couldn't read list of sources"));
-
-  pkgDPkgPM pm(*apt_cache_file);
-  pm.GetArchives(&fetcher, &l, apt_package_records);
-
-  f.FetchBytes=fetcher.FetchNeeded();
-  f.FetchPBytes=fetcher.PartialPresent();
-  f.DebBytes=fetcher.TotalNeeded();
-
-  return true;
-}
-
 static string reason_string_list(set<reason> &reasons)
 {
   set<reason>::iterator prev=reasons.end();
@@ -211,6 +181,11 @@ namespace
       mode = show_requiring_packages;
     summarize_reasons(reasons, mode, reason_strings);
 
+    // it seems that sometimes the code reaches here despite having nothing to
+    // show -- see #576584
+    if (reason_strings.empty())
+      return "";
+
     std::string rval = "(";
     if(mode == show_requiring_packages)
       rval += "for ";
@@ -432,7 +407,7 @@ string prompt_string(const string &prompt)
  *  fat warning message about them.  Returns false if the user doesn't
  *  want to continue.
  */
-static bool prompt_essential(const std::shared_ptr<terminal_metrics> &term_metrics)
+static bool prompt_essential(bool simulate_only, const std::shared_ptr<terminal_metrics> &term_metrics)
 {
   pkgvector todelete, whatsbroken;
   bool ok=true;
@@ -477,21 +452,39 @@ static bool prompt_essential(const std::shared_ptr<terminal_metrics> &term_metri
     {
       printf(_("WARNING: Performing this action will probably cause your system to break!\n         Do NOT continue unless you know EXACTLY what you are doing!\n"));
 
-      string untranslated_prompt = N_("I am aware that this is a very bad idea");
-      string prompt = _(untranslated_prompt.c_str());
-      char buf[1024];
+      if (simulate_only)
+	{
+	  printf("\n");
+	  printf(_("Ignoring confirmation in Simulation mode (subsequent actions will not have effects)\n"));
+	  printf(_("Press any key to continue...\n"));
+	  cin.get();
+	  return true;
+	}
+      else
+	{
+	  string untranslated_prompt = N_("I am aware that this is a very bad idea");
+	  string prompt = _(untranslated_prompt.c_str());
+	  char buf[1024];
 
-      printf(_("To continue, type the phrase \"%s\":\n"), prompt.c_str());
-      cin.getline(buf, 1023);
-      bool rval = (prompt == buf || untranslated_prompt == buf);
+	  if (prompt == untranslated_prompt)
+	    {
+	      printf(_("To continue, type the phrase \"%s\":\n"), prompt.c_str());
+	    }
+	  else
+	    {
+	      printf(_("To continue, type the phrase \"%s\" (or: \"%s\"):\n"), prompt.c_str(), untranslated_prompt.c_str());
+	    }
+	  cin.getline(buf, 1023);
+	  bool rval = (prompt == buf || untranslated_prompt == buf);
 
-      while(!cin && !cin.eof())
-	cin.getline(buf, 1023);
+	  while(!cin && !cin.eof())
+	    cin.getline(buf, 1023);
 
-      if(!cin)
-	throw StdinEOFException();
+	  if(!cin)
+	    throw StdinEOFException();
 
-      return rval;
+	  return rval;
+	}
     }
 
   return true;
@@ -503,7 +496,7 @@ static bool prompt_essential(const std::shared_ptr<terminal_metrics> &term_metri
  */
 static bool prompt_trust(const std::shared_ptr<terminal_metrics> &term_metrics)
 {
-  pkgvector untrusted;
+  std::vector<pkgCache::VerIterator> untrusted_versions;
 
   for(pkgCache::PkgIterator pkg=(*apt_cache_file)->PkgBegin();
       !pkg.end(); ++pkg)
@@ -517,18 +510,25 @@ static bool prompt_trust(const std::shared_ptr<terminal_metrics> &term_metrics)
 
 	  if((curr.end() || package_trusted(curr)) &&
 	     !package_trusted(cand))
-	    untrusted.push_back(pkg);
+	    {
+	      untrusted_versions.push_back(cand);
+	    }
 	}
     }
 
-  if(!untrusted.empty())
+  if(!untrusted_versions.empty())
     {
       printf(_("WARNING: untrusted versions of the following packages will be installed!\n\n"
 	       "Untrusted packages could compromise your system's security.\n"
 	       "You should only proceed with the installation if you are certain that\n"
 	       "this is what you want to do.\n\n"));
 
-      cmdline_show_pkglist(untrusted, term_metrics);
+      for (auto it : untrusted_versions)
+	{
+	  printf("  %s %s\n",
+		 it.ParentPkg().FullName(true).c_str(),
+		 get_uri(it, apt_package_records).c_str());
+	}
 
       printf("\n");
 
@@ -555,25 +555,36 @@ static bool prompt_trust(const std::shared_ptr<terminal_metrics> &term_metrics)
       // input methods.  Please include nothing but ASCII characters.
       // The text preceding the pipe character (|) will be ignored and
       // can be removed from your translation.
-      const string okstr    = P_("Go ahead and ignore the warning|Yes");
+      const string okstr    = P_("Go ahead and ignore the warning|yes");
       // TRANSLATORS: This string is a confirmation message, which
       // users (especially CJK users) should be able to input without
       // input methods.  Please include nothing but ASCII characters.
       // The text preceding the pipe character (|) will be ignored and
       // can be removed from your translation.
-      const string abortstr = P_("Abort instead of overriding the warning|No");
+      const string abortstr = P_("Abort instead of overriding the warning|no");
 
       // These strings are used to compare in a translation-invariant
       // way, so that "yes" and "no" are always valid inputs; if the
       // user can't enter the translated string for some reason,
       // he/she can always enter the fallback strings.
-      const string fallback_okstr = "Yes";
-      const string fallback_abortstr = "No";
+      const string fallback_okstr = "yes";
+      const string fallback_abortstr = "no";
 
-      while(1)
+      while (true)
 	{
 	  printf(_("Do you want to ignore this warning and proceed anyway?\n"));
-	  printf(_("To continue, enter \"%s\"; to abort, enter \"%s\": "), okstr.c_str(), abortstr.c_str());
+	  if (okstr == fallback_okstr && abortstr == fallback_abortstr)
+	    {
+	      printf(_("To continue, enter \"%s\"; to abort, enter \"%s\": "), okstr.c_str(), abortstr.c_str());
+	    }
+	  else
+	    {
+	      printf(_("To continue, enter \"%s\" (or \"%s\"); to abort, enter \"%s\" (or \"%s\"): "),
+		     okstr.c_str(),
+		     fallback_okstr.c_str(),
+		     abortstr.c_str(),
+		     fallback_abortstr.c_str());
+	    }
 	  char buf[1024];
 	  cin.getline(buf, 1023);
 	  buf[1023]='\0';
@@ -581,16 +592,17 @@ static bool prompt_trust(const std::shared_ptr<terminal_metrics> &term_metrics)
 	  if(cin.eof())
 	    throw StdinEOFException();
 
-
 	  const bool is_ok =             strncasecmp(okstr.c_str(), buf, okstr.size()) == 0;
 	  const bool is_fallback_ok =    strncasecmp(fallback_okstr.c_str(), buf, fallback_okstr.size()) == 0;
 	  const bool is_abort =          strncasecmp(abortstr.c_str(), buf, abortstr.size()) == 0;
 	  const bool is_fallback_abort = strncasecmp(fallback_abortstr.c_str(), buf, fallback_abortstr.size()) == 0;
+	  const bool is_rpmatch_ok      = (rpmatch(buf) > 0);
+	  const bool is_rpmatch_invalid = (rpmatch(buf) < 0);
 
-	  const bool rval = is_ok || (is_fallback_ok && !is_abort);
+	  const bool rval = is_ok || (is_fallback_ok && !is_abort) || is_rpmatch_ok;
 
-	  if(!is_ok && !is_abort && !is_fallback_ok && !is_fallback_abort)
-	    printf(_("Unrecognized input.  Enter either \"%s\" or \"%s\".\n"), okstr.c_str(), abortstr.c_str());
+	  if ((!is_ok && !is_abort && !is_fallback_ok && !is_fallback_abort) && is_rpmatch_invalid)
+	    printf(_("Unrecognized input\n"));
 	  else
 	    return rval;
 	}
@@ -615,7 +627,7 @@ bool cmdline_show_preview(bool as_upgrade, pkgset &to_install,
                           const std::shared_ptr<terminal_metrics> &term_metrics)
 {
   pkgvector lists[num_pkg_action_states];
-  pkgvector recommended, suggested;
+  pkgvector recommended, suggested, not_upgraded;
   pkgvector extra_install, extra_remove;
   unsigned long Upgrade=0, Downgrade=0, Install=0, ReInstall=0;
 
@@ -656,6 +668,11 @@ bool cmdline_show_preview(bool as_upgrade, pkgset &to_install,
 	      else if(package_suggested(pkg))
 		suggested.push_back(pkg);
 	    }
+	  break;
+	case pkg_hold:
+	case pkg_auto_hold:
+	  not_upgraded.push_back(pkg);
+	  break;
 	default:
 	  break;
 	}
@@ -712,6 +729,12 @@ bool cmdline_show_preview(bool as_upgrade, pkgset &to_install,
       cmdline_show_instinfo(suggested, verbose, showvers, showdeps, showsize, false, showwhy, term_metrics);
     }
 
+  if (verbose>0 && ! not_upgraded.empty())
+    {
+      printf(_("The following packages will NOT be UPGRADED:\n"));
+      cmdline_show_instinfo(not_upgraded, verbose, showvers, showdeps, showsize, false, showwhy, term_metrics);
+    }
+
   if((*apt_cache_file)->DelCount() == 0 &&
      (*apt_cache_file)->InstCount() == 0)
     printf(_("No packages will be installed, upgraded, or removed.\n"));
@@ -727,15 +750,15 @@ bool cmdline_show_preview(bool as_upgrade, pkgset &to_install,
   printf(_("%lu to remove and %lu not upgraded.\n"),
 	 (*apt_cache_file)->DelCount(),(*apt_cache_file)->KeepCount());
 
-  fetchinfo f;
-  if(get_fetchinfo(f))
+  std::unique_ptr<aptitude::apt::pkgAcquire_fetch_info> f = aptitude::apt::get_pkgAcquire_fetch_info();
+  if (f)
     {
-      if(f.DebBytes!=f.FetchBytes)
+      if (f->TotalNeeded != f->FetchNeeded)
 	printf(_("Need to get %sB/%sB of archives. "),
-	       SizeToStr(f.FetchBytes).c_str(), SizeToStr(f.DebBytes).c_str());
+	       SizeToStr(f->FetchNeeded).c_str(), SizeToStr(f->TotalNeeded).c_str());
       else
 	printf(_("Need to get %sB of archives. "),
-	       SizeToStr(f.DebBytes).c_str());
+	       SizeToStr(f->TotalNeeded).c_str());
     }
   else
     _error->DumpErrors();
@@ -747,14 +770,9 @@ bool cmdline_show_preview(bool as_upgrade, pkgset &to_install,
     printf(_("After unpacking %sB will be freed.\n"),
 	   SizeToStr(-(*apt_cache_file)->UsrSize()).c_str());
 
-  // If I return directly below, g++ complains about control reaching the
-  // end of a non-void function!
-  bool rval;
 
-  rval=((as_upgrade && !lists[pkg_upgrade].empty()) ||
-	!(extra_install.empty() && extra_remove.empty()));
-
-  return rval;
+  return ((as_upgrade && !lists[pkg_upgrade].empty()) ||
+	  !(extra_install.empty() && extra_remove.empty()));
 }
 
 static void cmdline_parse_show(string response,
@@ -772,7 +790,12 @@ static void cmdline_parse_show(string response,
 	it != packages.end(); ++it)
       do_cmdline_show(*it, verbose, term_metrics);
 
-  prompt_string(_("Press Return to continue."));
+  try {
+    prompt_string(_("Press Return to continue."));
+  }
+  catch (...) {
+    // ignore all exceptions
+  }
 }
 
 // Erm.  Merge w/ above?
@@ -785,9 +808,20 @@ static void cmdline_parse_changelog(string response, const std::shared_ptr<termi
   if(packages.empty())
     printf(_("No packages found -- enter the package names on the line after 'c'.\n"));
   else
-    do_cmdline_changelog(packages, term_metrics);
+    {
+      do_cmdline_changelog(packages, term_metrics);
+      if (_error->PendingError())
+	{
+	  _error->DumpErrors();
+	}
+    }
 
-  prompt_string(_("Press Return to continue."));
+  try {
+    prompt_string(_("Press Return to continue."));
+  }
+  catch (...) {
+    // ignore all exceptions
+  }
 }
 
 static void cmdline_parse_why(string response,
@@ -797,8 +831,11 @@ static void cmdline_parse_why(string response,
   // assume response[0]=='w'
   splitws(response, arguments, 1, response.size());
 
-  if(arguments.empty())
-    printf(_("No packages found -- enter zero or more roots of the search followed by the package to justify.\n"));
+  if (arguments.empty())
+    {
+      printf(_("No packages found -- enter zero or more roots of the search followed by the package to justify.\n"));
+      printf("\n");
+    }
   else
     {
       bool success;
@@ -811,12 +848,14 @@ static void cmdline_parse_why(string response,
 						false, false,
 						callbacks,
 						success));
-      if(frag.get() != NULL)
+      if (frag.get())
         {
           const unsigned int screen_width = term_metrics->get_screen_width();
           std::cout << frag->layout(screen_width, screen_width, cwidget::style());
         }
+
       _error->DumpErrors();
+      std::cout << std::endl;
     }
 }
 
@@ -935,6 +974,8 @@ bool cmdline_do_prompt(bool as_upgrade,
 		       bool force_no_change,
 		       pkgPolicy &policy,
 		       bool arch_only,
+		       bool download_only,
+		       bool simulate_only,
                        const std::shared_ptr<terminal_metrics> &term_metrics)
 {
   bool exit=false;
@@ -1043,18 +1084,38 @@ bool cmdline_do_prompt(bool as_upgrade,
 		}
 	    }
 
+	  if (download_only)
+	    {
+	      printf("\n");
+	      printf(_("Note: Using 'Download Only' mode, no other actions will be performed.\n"));
+	    }
+
 	  while(!valid_response)
 	    {
 	      valid_response=true;
 	      fflush(stdout);
 
+	      if (simulate_only)
+		{
+		  printf("\n");
+		  printf(_("Note: Using 'Simulate' mode.\n"));
+		}
+
 	      string prompt =
 		!have_broken
 		    ? _("Do you want to continue? [Y/n/?] ")
 		    : _("Resolve these dependencies by hand? [N/+/-/_/:/?] ");
 
-	      string response=prompt_string(prompt);
-	      string::size_type loc=0;
+	      string response;
+	      string::size_type loc = 0;
+	      try {
+		response = prompt_string(prompt);
+	      }
+	      catch (...) {
+		// ignore all exceptions, default to "no"
+		response = "n";
+		printf(_("%s (stdin unavailable)\n"), response.c_str());
+	      }
 
 	      while(loc<response.size() && isspace(response[loc]))
 		++loc;
@@ -1154,7 +1215,14 @@ bool cmdline_do_prompt(bool as_upgrade,
 		  }
 		  break;
 		case 'E':
-		  ui_preview();
+		  {
+		    // prevent curses if dumb term -- see #317928, #817276
+		    if (aptitude::util::is_dumb_terminal())
+		      aptitude::util::print_ncurses_dumb_terminal();
+		    else
+		      ui_preview();
+		  }
+		  break;
 		case '?':
 		  valid_response=false;
 		  prompt_help(cout, have_broken, term_metrics);
@@ -1168,7 +1236,7 @@ bool cmdline_do_prompt(bool as_upgrade,
 	}
 
       // Note: only show the prompt if we're planning to continue.
-      if(rval && (!prompt_essential(term_metrics) || !prompt_trust(term_metrics)))
+      if(rval && (!prompt_essential(simulate_only, term_metrics) || !prompt_trust(term_metrics)))
 	{
 	  rval=false;
 	  exit=true;
