@@ -1,6 +1,7 @@
 // cmdline_prompt.h                     -*-c++-*-
 //
 // Copyright (C) 2004, 2010 Daniel Burrows
+// Copyright (C) 2014-2016 Manuel A. Fernandez Montecelo
 //
 // This program is free software; you can redistribute it and/or
 // modify it under the terms of the GNU General Public License as
@@ -14,8 +15,8 @@
 //
 // You should have received a copy of the GNU General Public License
 // along with this program; see the file COPYING.  If not, write to
-// the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-// Boston, MA 02111-1307, USA.
+// the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
+// Boston, MA 02110-1301, USA.
 
 #ifndef CMDLINE_PROMPT_H
 #define CMDLINE_PROMPT_H
@@ -81,6 +82,8 @@ bool cmdline_do_prompt(bool as_upgrade,
 		       bool force_no_change,
 		       pkgPolicy &policy,
 		       bool arch_only,
+		       bool download_only,
+		       bool simulate_only,
                        const std::shared_ptr<aptitude::cmdline::terminal_metrics> &term_metrics);
 
 bool cmdline_show_preview(bool as_upgrade, pkgset &to_install,
