@@ -1,7 +1,7 @@
 // apt.h  -*-c++-*-
 //
 //  Copyright 1999-2002, 2004-2005, 2007-2010 Daniel Burrows
-//  Copyright 2015 Manuel A. Fernandez Montecelo
+//  Copyright 2015-2016 Manuel A. Fernandez Montecelo
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -15,8 +15,8 @@
 //
 //  You should have received a copy of the GNU General Public License
 //  along with this program; see the file COPYING.  If not, write to
-//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-//  Boston, MA 02111-1307, USA.
+//  the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
+//  Boston, MA 02110-1301, USA.
 //
 
 
@@ -62,13 +62,16 @@ extern aptitudeCacheFile *apt_cache_file;
 extern resolver_manager *resman;
 extern pkgSourceList *apt_source_list;
 extern pkgRecords *apt_package_records;
+extern bool shutdown_in_progress;
 
 void apt_preinit(const char *rootdir);
 // Performs initialization of stuff that has to happen before apt_init is
 // called (eg, pkgInitialize and setting up the undo structure)
 
 void apt_init(OpProgress *progess_bar,
-	      bool do_initselections, const char * status_fname=NULL);
+	      bool do_initselections,
+	      bool operation_needs_lock,
+	      const char * status_fname);
 //  It actually doesn't do what you expect!  This routine is a NOP if
 // it has already been called; the rationale is that it allows any apt-using
 // class to call this in its constructor (thus guaranteeing that the cache
@@ -105,12 +108,18 @@ bool get_apt_knows_about_rootdir();
  *                           sticky database.
  *  \param status_fname if not \b NULL, a filename to use in lieu
  *                      of /var/lib/aptitude/pkgstates.
+ *  \param reset_reinstall Reset packages set for reinstall (do this after successful installation only).
  */
 void apt_load_cache(OpProgress *progress_bar,
-		    bool do_initselections, const char *status_fname = NULL);
+		    bool do_initselections,
+		    bool operation_needs_lock,
+		    const char *status_fname,
+		    bool reset_reinstall);
 
 void apt_reload_cache(OpProgress *progress_bar,
-		      bool do_initselections, const char * status_fname=NULL);
+		      bool do_initselections,
+		      bool operation_needs_lock,
+		      const char * status_fname);
 //  Forces the cache to be reloaded.
 //
 //  NOTE: at the moment, the interface won't work too well while this
@@ -138,7 +147,7 @@ extern sigc::signal0<void> consume_errors;
 /** \brief Used to cache downloaded data, to avoid multiple
  *  downloads of items such as changelogs and screenshots.
  */
-extern std::shared_ptr<aptitude::util::file_cache> download_cache;
+std::shared_ptr<aptitude::util::file_cache> get_download_cache();
 
 void apt_dumpcfg(const char *root);
 // Dumps a subtree of the configuration to ~/.aptitude/config
@@ -221,6 +230,34 @@ void surrounding_or(pkgCache::DepIterator dep,
 		    pkgCache *cache = NULL);
 
 
+/** Get URI of a package version
+ *
+ * @param ver The package version
+ * @param records The package records
+ *
+ * @return The requested value
+ */
+std::string get_uri(const pkgCache::VerIterator& ver,
+		    const pkgRecords* records);
+
+/** Get Origin of a package version
+ *
+ * @param ver The package version
+ * @param records The package records
+ *
+ * @return The requested value
+ */
+std::string get_origin(const pkgCache::VerIterator& ver,
+		       const pkgRecords* records);
+
+/** Get candidate version (iterator) of a package (iterator)
+ *
+ * @param pkg The package iterator
+ *
+ * @return The requested value
+ */
+pkgCache::VerIterator get_candidate_version(const pkgCache::PkgIterator& pkg);
+
 /** \return a short description string corresponding to the given
  *  version.
  */
@@ -273,16 +310,16 @@ pkgCache::DepIterator is_conflicted(const pkgCache::VerIterator &ver,
  *
  * @cache Dependency cache
  *
- * @param follow_recommends Whether to follow recommends
+ * @param keep_recommends_installed Whether to keep recommends
  *
- * @param follow_suggests Whether to follow suggests
+ * @param keep_suggests_installed Whether to keep suggests
  *
  * @return Whether the action is allowed or not.
  */
 bool can_remove_autoinstalled(const pkgCache::PkgIterator& pkg,
 			      aptitudeDepCache& cache,
-			      bool follow_recommends,
-			      bool follow_suggests);
+			      bool keep_recommends_installed,
+			      bool keep_suggests_installed);
 
 /** Check if package is virtual
  *
@@ -320,6 +357,17 @@ inline bool is_auto_installed(const pkgDepCache::StateCache& state)
   return (state.Flags & pkgCache::Flag::Auto);
 }
 
+/** Check if package is auto installed
+ *
+ * @param pkg Package to check
+ *
+ * @return Whether the package is auto installed or not
+ */
+inline bool is_auto_installed(const pkgCache::PkgIterator& pkg)
+{
+  return (apt_cache_file && is_auto_installed((*apt_cache_file)[pkg]));
+}
+
 /** A pair (veriterator,verfile) -- used for building a list of
  *  versions sorted by file location.
  */
@@ -328,15 +376,26 @@ typedef std::pair<pkgCache::VerIterator, pkgCache::VerFileIterator> loc_pair;
 /** Used to compare two version files based on their location. */
 struct location_compare
 {
-  bool operator()(loc_pair a, loc_pair b) const
+  inline bool operator()(const loc_pair& a, const loc_pair& b) const
   {
-    if(a.second->File == b.second->File)
-      return a.second->Offset<b.second->Offset;
+    if (a.second->File == b.second->File)
+      return (a.second->Offset < b.second->Offset);
     else
-      return a.second->File<b.second->File;
+      return (a.second->File < b.second->File);
   }
 };
 
+
+/** Whether a particular version is security-related
+ *
+ * Useful e.g. to classify in menus as "Security Upgrade"
+ *
+ * @return \b true iff the given package version comes from security.d.o or
+ * known places
+ */
+bool is_security(const pkgCache::VerIterator &ver);
+
+
 /** \return \b true if the given dependency is "interesting":
  *          specifically, if it's either critical or a Recommends
  *          that's "new" or currently satisfied.
@@ -477,6 +536,17 @@ namespace aptitude
 {
   namespace apt
   {
+    /** Translate from Priority type in apt to a string
+     *
+     * @param priority Priority
+     *
+     * @param short_form Whether to return three letter "code" instead of full
+     * name
+     *
+     * @return Priority as string
+     */
+    std::string priority_to_string(const pkgCache::State::VerPriority priority, bool short_form = false);
+
     /** \return \b true if the given dependency is a Replaces dependency
      *  and participates in a conflicts/provides/replaces relationship.
      *
@@ -501,6 +571,31 @@ namespace aptitude
     {
       return !is_native_arch(ver) && (strcmp(ver.Arch(), "all") != 0);
     }
+
+    /** Clean cache of downloaded packages from the canonical directory
+     *
+     * @return Whether the operation was successful
+     */
+    bool clean_cache_dir();
+
+
+    /** pkgAcquire fetch information
+     */
+    struct pkgAcquire_fetch_info {
+      unsigned long long FetchNeeded = 0, PartialPresent = 0, TotalNeeded = 0;
+
+      pkgAcquire_fetch_info() { }
+      pkgAcquire_fetch_info(unsigned long long F,
+			    unsigned long long P,
+			    unsigned long long T) :
+	FetchNeeded { F }, PartialPresent { P }, TotalNeeded { T }
+      { }
+    };
+    /** Get pkgAcquire fetch information
+     *
+     * @return Information requested, that can be nullptr if error
+     */
+    std::unique_ptr<pkgAcquire_fetch_info> get_pkgAcquire_fetch_info();
   }
 }
 
