@@ -1,7 +1,7 @@
 // aptcache.cc
 //
 //  Copyright 1999-2009, 2011 Daniel Burrows
-//  Copyright 2015 Manuel A. Fernandez Montecelo
+//  Copyright 2015-2016 Manuel A. Fernandez Montecelo
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -15,8 +15,8 @@
 //
 //  You should have received a copy of the GNU General Public License
 //  along with this program; see the file COPYING.  If not, write to
-//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-//  Boston, MA 02111-1307, USA.
+//  the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
+//  Boston, MA 02110-1301, USA.
 
 #include "aptcache.h"
 
@@ -220,9 +220,9 @@ aptitudeDepCache::aptitudeDepCache(pkgCache *Cache, Policy *Plcy)
 #endif
 }
 
-bool aptitudeDepCache::Init(OpProgress *Prog, bool WithLock, bool do_initselections, const char *status_fname)
+bool aptitudeDepCache::Init(OpProgress *Prog, bool WithLock, bool do_initselections, const char *status_fname, bool reset_reinstall)
 {
-  return build_selection_list(*Prog, WithLock, do_initselections, status_fname);
+  return build_selection_list(Prog, WithLock, do_initselections, status_fname, reset_reinstall);
 }
 
 aptitudeDepCache::~aptitudeDepCache()
@@ -239,104 +239,18 @@ void aptitudeDepCache::set_read_only(bool new_read_only)
   read_only = new_read_only;
 }
 
-namespace
-{
-  // User tag syntax:
-  // 
-  // Tag-List ::= Tag+
-  // Tag      ::= regexp([^[:space:]]) | regexp("([^\\"]|\\.)")
-  bool parse_user_tag(std::string &out,
-		      const char *&start, const char *end,
-		      const std::string &package_name)
-  {
-    const char * const initial_start = start;
-    while(start != end && isspace(*start))
-      ++start;
-
-    if(start == end)
-      return false;
-    else if(*start != '"')
-      {
-	while(start != end && !isspace(*start))
-	  {
-	    out += *start;
-	    ++start;
-	  }
-	return true;
-      }
-    else
-      {
-	++start;
-	while(start != end && *start != '"')
-	  {
-	    if(*start == '\\')
-	      {
-		++start;
-		if(start == end)
-		  return _error->Error(_("Error parsing a user-tag for the package %s: unexpected end-of-line following %s."),
-				       package_name.c_str(),
-				       (std::string("\"") + std::string(initial_start, start)).c_str());
-		else
-		  {
-		    out += *start;
-		    ++start;
-		  }
-	      }
-	    else
-	      {
-		out += *start;
-		++start;
-	      }
-	  }
-
-	if(start == end)
-	  return _error->Error(_("Unterminated '\"' in the user-tags list of the package %s."),
-			       package_name.c_str());
-	else
-	  {
-	    ++start;
-	    return true;
-	  }
-      }
-  }
-}
-
-void aptitudeDepCache::parse_user_tags(std::set<user_tag> &tags,
-				       const char *&start, const char *end,
-				       const std::string &package_name)
-{
-  while(start != end)
-    {
-      while(start != end && isspace(*start))
-	++start;
-
-      std::string tag;
-      parse_user_tag(tag, start, end, package_name);
-
-      typedef std::map<std::string, user_tag_reference>::const_iterator
-	user_tags_index_iterator;
-      user_tags_index_iterator found = user_tags_index.find(tag);
-      if(found == user_tags_index.end())
-	{
-	  user_tag_reference loc(user_tags.size());
-	  user_tags.push_back(tag);
-	  std::pair<user_tags_index_iterator, bool> tmp(user_tags_index.insert(std::make_pair(tag, loc)));
-	  found = tmp.first;
-	}
-      tags.insert(user_tag(found->second));
-    }
-}
-
-bool aptitudeDepCache::build_selection_list(OpProgress &Prog, bool WithLock,
+bool aptitudeDepCache::build_selection_list(OpProgress* Prog,
+					    bool WithLock,
 					    bool do_initselections,
-					    const char *status_fname)
+					    const char* status_fname,
+					    bool reset_reinstall)
 {
   action_group group(*this);
 
   bool initial_open=false;
   // This will be set to true if the state file does not exist.
 
-  if(!pkgDepCache::Init(&Prog))
+  if(!pkgDepCache::Init(Prog))
     return false;
 
   records = new pkgRecords(*this);
@@ -347,8 +261,6 @@ bool aptitudeDepCache::build_selection_list(OpProgress &Prog, bool WithLock,
   // Garbage set up (they behave more the way they ought to then).
   MarkAndSweep();
 
-  string statedir=aptcfg->FindDir("Dir::Aptitude::state", STATEDIR);
-  // Should this not go under Dir:: ?  I'm not sure..
   delete package_states;
   package_states=new aptitude_state[Head().PackageCount];
   user_tags.clear();
@@ -380,13 +292,14 @@ bool aptitudeDepCache::build_selection_list(OpProgress &Prog, bool WithLock,
   // manages to trigger a mark operation.
   duplicate_cache(&backup_state);
 
+  // Should this not go under Dir:: ?  I'm not sure..
+  string statedir = aptcfg->FindDir("Dir::Aptitude::state", STATEDIR);
+  string statefilepath = (status_fname) ? status_fname : (statedir + "/" + "pkgstates");
+
   FileFd state_file;
 
   // Read in the states that we saved
-  if(status_fname==NULL)
-    state_file.Open(statedir+"pkgstates", FileFd::ReadOnly);
-  else
-    state_file.Open(status_fname, FileFd::ReadOnly);
+  state_file.Open(statefilepath, FileFd::ReadOnly);
 
   // Have to make the file NOT read-only to set up the initial state.
   read_only = false;
@@ -408,12 +321,18 @@ bool aptitudeDepCache::build_selection_list(OpProgress &Prog, bool WithLock,
     }
   else
     {
-      int file_size=state_file.Size();
-      Prog.OverallProgress(0, file_size, 1, _("Reading extended state information"));
+      // last percent shown in progress -- do not update on every cycle
+      int last_pct_shown = 0;
+      int amt = 0;
+      int file_size = state_file.Size();
+      if (Prog)
+	{
+	  Prog->OverallProgress(0, file_size, 1, _("Reading extended state information"));
+	}
 
       pkgTagFile tagfile(&state_file);
       pkgTagSection section;
-      int amt=0;
+
       bool do_dselect=aptcfg->FindB(PACKAGE "::Track-Dselect-State", true);
       while(tagfile.Step(section))
 	{
@@ -443,6 +362,24 @@ bool aptitudeDepCache::build_selection_list(OpProgress &Prog, bool WithLock,
 	      section.FindFlag("Upgrade", tmp, 1);
 	      pkg_state.upgrade=(tmp==1);
 
+	      tmp=0;
+	      section.FindFlag("Reinstall", tmp, 1);
+	      pkg_state.reinstall = (tmp==1);
+
+	      // if the last installation was successful reset_reinstall==true,
+	      // to unmark .reinstall property of the package, otherwise there
+	      // is no way to control that this is not repeated forever.
+	      //
+	      // doing the more expensive test "(reset_reinstall &&
+	      // pkg_state.reinstall)" rather than blindly marking as false,
+	      // because of "dirty" (so we only mark as dirty and pending to
+	      // save when we really changed the state.
+	      if (reset_reinstall && pkg_state.reinstall)
+		{
+		  pkg_state.reinstall = false;
+		  dirty = true;
+		}
+
 	      unsigned long auto_new_install = 0;
 	      section.FindFlag("Auto-New-Install", auto_new_install, 1);
 	      if(auto_new_install)
@@ -473,9 +410,16 @@ bool aptitudeDepCache::build_selection_list(OpProgress &Prog, bool WithLock,
 
 	      {
 		const char *start, *end;
-		if(section.Find("User-Tags", start, end))
-		  parse_user_tags(pkg_state.user_tags, start, end,
-				  package_name);
+		if (section.Find("User-Tags", start, end))
+		  {
+		    bool parse_ok = user_tags.parse(pkg_state.user_tags, start, end,
+						    package_name);
+		    if (!parse_ok)
+		      {
+			_error->Error(_("Cannot parse user-tags for package: %s: '%s'"), pkg.FullName(true).c_str(), string(start, end).c_str());
+			// do not return now, loading would hang
+		      }
+		  }
 	      }
 
 	      if(do_dselect && pkg->SelectedState != last_dselect_state)
@@ -518,24 +462,61 @@ bool aptitudeDepCache::build_selection_list(OpProgress &Prog, bool WithLock,
 	      // "downgrade") again down in this function, and while for
 	      // "upgrades" it is not a problem, with "downgrades" it will not
 	      // show as upgradable in the current interactive session.
+	      //
+	      //
+	      // also unmark as upgrade as soon as already upgraded, even if no
+	      // candidate version required in pkgstates -- see #721426
 	      std::string installed_ver = pkg.CurVersion() ? pkg.CurVersion() : "";
-	      if (pkg_state.upgrade && installed_ver == candver)
+	      if (pkg_state.upgrade)
 		{
-		  pkg_state.upgrade = false;
-		  pkg_state.candver = "";
-		  dirty = true;
+		  bool version_as_in_pkgstates = (!pkg_state.candver.empty() && (installed_ver == pkg_state.candver));
+		  bool version_as_candidate = (!GetCandidateVersion(pkg).end() && !pkg.CurrentVer().end() && (GetCandidateVersion(pkg) == pkg.CurrentVer()));
+		  if (version_as_in_pkgstates || (pkg_state.candver.empty() && version_as_candidate))
+		    {
+		      pkg_state.upgrade = false;
+		      pkg_state.candver = "";
+		      dirty = true;
+		    }
+		}
+	    }
+
+	  if (Prog)
+	    {
+	      // update progress, but not every time -- very expensive
+	      amt += section.size();
+	      int pct = (file_size > 0) ? (100*amt) / file_size : 0;
+	      if ((pct % 10 == 1) && last_pct_shown != pct)
+		{
+		  last_pct_shown = pct;
+		  Prog->OverallProgress(amt, file_size, 1, _("Reading extended state information"));
 		}
 	    }
-	  amt+=section.size();
-	  Prog.OverallProgress(amt, file_size, 1, _("Reading extended state information"));
 	}
-      Prog.OverallProgress(file_size, file_size, 1, _("Reading extended state information"));
-      Prog.Done();
+
+      // if pkgTagFile.Step() throws errors, file likely corrupt -- see #405506
+      if (_error->PendingError())
+	{
+	  _error->Error(_("Problem parsing '%s', is it corrupt or malformed? You can try to recover from '%s.old'."),
+			statefilepath.c_str(), statefilepath.c_str());
+	  return false;
+	}
+
+      if (Prog)
+	{
+	  Prog->OverallProgress(file_size, file_size, 1, _("Reading extended state information"));
+	  Prog->Done();
+	}
     }
 
-  int num=0;
-
-  Prog.OverallProgress(0, Head().PackageCount, 1, _("Initializing package states"));
+  int progress_num = 0;
+  int progress_total = Head().PackageCount;
+  // only update if we're going to increase 10% or so, minimum 1 (to avoid
+  // divide by zero)
+  int update_progress_10pct = std::max(progress_total / 10, 1);
+  if (Prog)
+    {
+      Prog->OverallProgress(0, progress_total, 1, _("Initializing package states"));
+    }
 
   new_package_count=0;
 
@@ -618,22 +599,34 @@ bool aptitudeDepCache::build_selection_list(OpProgress &Prog, bool WithLock,
 	  dirty = true;
 	}
 
-      ++num;
-      Prog.OverallProgress(num, Head().PackageCount, 1, _("Initializing package states"));
+      if (Prog)
+	{
+	  // don't update on every cycle
+	  if ((++progress_num % update_progress_10pct) == 1)
+	    {
+	      Prog->OverallProgress(progress_num, progress_total, 1, _("Initializing package states"));
+	    }
+	}
     }
 
-  Prog.OverallProgress(Head().PackageCount, Head().PackageCount, 1, _("Initializing package states"));
+  if (Prog)
+    Prog->OverallProgress(progress_total, progress_total, 1, _("Initializing package states"));
 
   duplicate_cache(&backup_state);
 
   if(aptcfg->FindB(PACKAGE "::Auto-Upgrade", false) && do_initselections)
     mark_all_upgradable(aptcfg->FindB(PACKAGE "::Auto-Install", true),
 			true, NULL);
-  Prog.Done();
+
+  if (Prog)
+    Prog->Done();
 
   read_only = (lock == -1);
 
-  return true;
+  if (_error->PendingError())
+    return false;
+  else
+    return true;
 }
 
 void aptitudeDepCache::mark_all_upgradable(bool with_autoinst,
@@ -692,7 +685,6 @@ void aptitudeDepCache::get_upgradable(bool ignore_removed,
 
   for(pkgCache::PkgIterator p = PkgBegin(); !p.end(); ++p)
     {
-      StateCache &state = (*this)[p];
       aptitude_state &estate = get_ext_state(p);
 
       if(p.CurrentVer().end())
@@ -705,7 +697,9 @@ void aptitudeDepCache::get_upgradable(bool ignore_removed,
 
       if(!ignore_removed)
 	{
-	  do_upgrade = state.Status > 0 && !is_held(p);
+	  // allow downgrades when pinned high -- see #344700, #348679
+	  do_upgrade = !is_held(p) && !GetCandidateVersion(p).end() && (GetCandidateVersion(p) != p.CurrentVer());
+
 	  if(do_upgrade)
 	    LOG_DEBUG(logger, p.FullName(false) << " is upgradable.");
 	  else
@@ -722,7 +716,8 @@ void aptitudeDepCache::get_upgradable(bool ignore_removed,
 
 	      // Fall through
 	    case pkgCache::State::Install:
-	      if(state.Status > 0 && !is_held(p))
+	      // allow downgrades when pinned high -- see #344700, #348679
+	      if (!is_held(p) && !GetCandidateVersion(p).end() && (GetCandidateVersion(p) != p.CurrentVer()))
 		{
 		  do_upgrade = true;
 		  LOG_TRACE(logger, p.FullName(false) << " is upgradable.");
@@ -746,8 +741,8 @@ void aptitudeDepCache::get_upgradable(bool ignore_removed,
 //  FIXME: clean up the logic by having an internal "write to this fd"
 // routine and an exported "ok, set up for the write and then clean up"
 // routine.
-bool aptitudeDepCache::save_selection_list(OpProgress &prog,
-					   const char *status_fname)
+bool aptitudeDepCache::save_selection_list(OpProgress* Prog,
+					   const char* status_fname)
 {
   // Refuse to write to disk if nothing changed and we aren't writing
   // to an unusual file
@@ -764,7 +759,7 @@ bool aptitudeDepCache::save_selection_list(OpProgress &prog,
   // Ow, that sucks.  The solution will be to write the apt states to
   // a separate file.
   if(status_fname == NULL)
-    writeStateFile(&prog);
+    writeStateFile(Prog, false);
 
   // helper class to save selection state of packages to dpkg database
   aptitude::apt::dpkg::DpkgSelections dpkg_selections;
@@ -783,94 +778,101 @@ bool aptitudeDepCache::save_selection_list(OpProgress &prog,
   fchmod(newstate.Fd(), 0644);
 
   if(!newstate.IsOpen())
-    _error->Error(_("Cannot open Aptitude state file"));
+    {
+      _error->Error(_("Cannot open Aptitude state file"));
+      if (Prog)
+	Prog->Done();
+      return false;
+    }
   else
     {
-      int num=0;
-      prog.OverallProgress(0, Head().PackageCount, 1, _("Writing extended state information"));
+      int progress_num = 0;
+      int progress_total = Head().PackageCount;
+      // only update if we're going to increase 10% or so, minimum 1 (to avoid
+      // divide by zero)
+      int update_progress_10pct = std::max(progress_total / 10, 1);
+      if (Prog)
+	{
+	  Prog->OverallProgress(0, progress_total, 1, _("Writing extended state information"));
+	}
+
+      // save some allocations in the loop and other optimisations -- see #312920
+      std::stringstream newstate_tmpbuffer;
+      string line;
+      string forbidstr;
+      string upgradestr;
+      string reinstall_str;
+      string autostr;
+      string tailstr;
+      string user_tags;
+      string select_arch;
 
       for(PkgIterator i=PkgBegin(); !i.end(); i++)
-	if(!i.VersionList().end())
+	if (i.VersionList().end())
+	  {
+	    ++progress_num;
+	  }
+	else
 	  {
 	    StateCache &state=(*this)[i];
 	    aptitude_state &estate=get_ext_state(i);
 
-	    string forbidstr=!estate.forbidver.empty()
-	      ? "ForbidVer: "+estate.forbidver+"\n":"";
+	    forbidstr = (!estate.forbidver.empty()) ? ("ForbidVer: " + estate.forbidver + "\n") : "";
 
-	    bool upgrade=(!i.CurrentVer().end()) && state.Install();
-	    string upgradestr=upgrade ? "Upgrade: yes\n" : "";
+	    upgradestr = ((!i.CurrentVer().end()) && state.Install()) ? "Upgrade: yes\n" : "";
 
+	    reinstall_str = (estate.reinstall) ? ("Reinstall: yes\n") : "";
+
+	    // packages that are auto and not yet installed are marked in this
+	    // way in aptitude's DB, to set the flag accordingly when installing
+	    // for the first time.  apt does not save auto state on uninstalled
+	    // packages.
+	    //
+	    // it comes originally from fix to #435079; and then to fix problem
+	    // when install-auto scheduled from previous sessions (but were not
+	    // installed at that point) were losing the auto flag (#563877).
+	    //
+	    // the flag is removed if the package is already installed, to fix
+	    // having this flag and preventing code handling auto in different
+	    // parts from working normally (#816497)
+	    bool is_or_was_auto = ((state.Flags & Flag::Auto) != 0) || estate.previously_auto_package;
 	    bool auto_new_install = (i.CurrentVer().end() &&
 				     state.Install() &&
-				     ((state.Flags & Flag::Auto) != 0));
-	    string autostr = auto_new_install ? "Auto-New-Install: yes\n" : "";
+				     is_or_was_auto);
+	    autostr = (auto_new_install) ? "Auto-New-Install: yes\n" : "";
 
-	    string tailstr;
 
 	    if(state.Install() &&
 	       !estate.candver.empty() &&
-	       (GetCandidateVer(i).end() ||
-		GetCandidateVer(i).VerStr() != estate.candver))
+	       (GetCandidateVersion(i).end() ||
+		GetCandidateVersion(i).VerStr() != estate.candver))
 	      tailstr = "Version: " + estate.candver + "\n";
+	    else
+	      tailstr = "";
 
 	    // Build the list of usertags for this package.
-	    std::string user_tags;
-	    if(!estate.user_tags.empty())
+	    if (!estate.user_tags.empty())
 	      {
-		user_tags = "User-Tags: ";
+		user_tags = "User-Tags:";
 
-		// Put the usertags in sorted order so we get
-		// predictable outputs.
-		std::vector<std::string> tmp;
-		tmp.reserve(estate.user_tags.size());
+		// get sorted usertags so we get predictable outputs
+		std::vector<std::string> sorted_pkg_user_tags = get_user_tags(i);
 
-		for(std::set<user_tag>::const_iterator it
-		      = estate.user_tags.begin(); it != estate.user_tags.end(); ++it)
-		  tmp.push_back(deref_user_tag(*it));
-
-		std::sort(tmp.begin(), tmp.end());
-
-		bool first = true;
-		// Append user tags to the field, using double-quotes
-		// if the tag contains spaces or double-quotes.
-		for(std::vector<std::string>::const_iterator it = tmp.begin();
-		    it != tmp.end(); ++it)
+		// append user tags to the field
+		for (const auto& tag : sorted_pkg_user_tags)
 		  {
-		    if(first)
-		      first = false;
-		    else
-		      user_tags.push_back(' ');
-
-		    if(it->find_first_of(" \"\\") != std::string::npos)
-		      {
-			user_tags.push_back('"');
-			for(std::string::const_iterator tag_it = it->begin();
-			    tag_it != it->end(); ++tag_it)
-			  {
-			    switch(*tag_it)
-			      {
-			      case '"':
-			      case '\\':
-				user_tags.push_back('\\');
-				user_tags.push_back(*tag_it);
-				break;
-			      default:
-				user_tags.push_back(*tag_it);
-				break;
-			      }
-			  }
-			user_tags.push_back('"');
-		      }
-		    else
-		      user_tags.insert(user_tags.size(), *it);
+		    user_tags.push_back(' ');
+		    user_tags += tag;
 		  }
 
 		user_tags.push_back('\n');
 	      }
+	    else
+	      user_tags = "";
 
+	    // write section for this package
 	    using cw::util::ssprintf;
-	    std::string line(ssprintf("Package: %s\nArchitecture: %s\nUnseen: %s\nState: %i\nDselect-State: %i\nRemove-Reason: %i\n%s%s%s%s%s\n",
+	    line = ssprintf("Package: %s\nArchitecture: %s\nUnseen: %s\nState: %i\nDselect-State: %i\nRemove-Reason: %i\n%s%s%s%s%s%s\n",
 				      i.Name(),
                                       i.Arch(),
 				      estate.new_package?"yes":"no",
@@ -878,10 +880,12 @@ bool aptitudeDepCache::save_selection_list(OpProgress &prog,
 				      i->SelectedState,
 				      estate.remove_reason,
 				      upgradestr.c_str(),
+				      reinstall_str.c_str(),
 				      autostr.c_str(),
 				      forbidstr.c_str(),
 				      user_tags.c_str(),
-				      tailstr.c_str()));
+				      tailstr.c_str());
+	    newstate_tmpbuffer << line;
 
 	    // dpkg-dselect state
 	    if (estate.original_selection_state != estate.selection_state
@@ -889,12 +893,12 @@ bool aptitudeDepCache::save_selection_list(OpProgress &prog,
 	      {
 		// for internal reasons, apt's PkgIterator is always arch:any,
 		// not arch:all.  VerIterator can be arch:all, though.
-		std::string select_arch = i.Arch();
+		select_arch = i.Arch();
 
 		// all install, downgrade, upgrade, etc -- try to use candidate version
-		if (state.Install() && !GetCandidateVer(i).end())
+		if (state.Install() && !GetCandidateVersion(i).end())
 		  {
-		    select_arch = GetCandidateVer(i).Arch();
+		    select_arch = GetCandidateVersion(i).Arch();
 		  }
 		// all delete, purge, hold, keep etc -- try to use current version
 		else if (! i.CurrentVer().end())
@@ -909,35 +913,34 @@ bool aptitudeDepCache::save_selection_list(OpProgress &prog,
 		estate.original_selection_state = estate.selection_state;
 	      }
 
-	    if(newstate.Failed() || !newstate.Write(line.c_str(), line.size()))
+	    if (Prog)
 	      {
-		_error->Error(_("Couldn't write state file"));
-		newstate.Close();
-
-		if(!status_fname)
-		  unlink((statefile+".new").c_str());
-		return false;
+		// don't update on every cycle
+		if ((++progress_num % update_progress_10pct) == 1)
+		  {
+		    Prog->OverallProgress(progress_num, progress_total, 1, _("Writing extended state information"));
+		  }
 	      }
-
-	    num++;
-	    prog.OverallProgress(num, Head().PackageCount, 1, _("Writing extended state information"));
 	  }
 
-      prog.OverallProgress(Head().PackageCount, Head().PackageCount, 1, _("Writing extended state information"));
+      if (Prog)
+	Prog->OverallProgress(progress_total, progress_total, 1, _("Writing extended state information"));
 
-      if(newstate.Failed())
-	// This is /probably/ redundant, but paranoia never hurts.
+      if (newstate.Failed() ||
+	  !newstate.Write(newstate_tmpbuffer.str().c_str(), newstate_tmpbuffer.str().size()))
 	{
-	  _error->Error(_("Error writing state file"));
+	  _error->Error(_("Couldn't write state file"));
 	  newstate.Close();
 
-	  if(!status_fname)
+	  if (!status_fname)
 	    unlink((statefile+".new").c_str());
 
-	  prog.Done();
+	  if (Prog)
+	    Prog->Done();
 	  return false;
 	}
       newstate.Close();
+
       // FIXME!  This potentially breaks badly on NFS.. (?) -- actually, it
       //       wouldn't be harmful; you'd just get gratuitous errors..
       if(!status_fname)
@@ -947,7 +950,8 @@ bool aptitudeDepCache::save_selection_list(OpProgress &prog,
 	  if(unlink(oldstr.c_str()) != 0 && errno != ENOENT)
 	    {
 	      _error->Errno("save_selection_list", _("failed to remove %s"), oldstr.c_str());
-	      prog.Done();
+	      if (Prog)
+		Prog->Done();
 	      return false;
 	    }
 
@@ -955,14 +959,16 @@ bool aptitudeDepCache::save_selection_list(OpProgress &prog,
 	    {
 	      _error->Errno("save_selection_list", _("failed to rename %s to %s"),
 			    statefile.c_str(), (statefile + ".old").c_str());
-	      prog.Done();
+	      if (Prog)
+		Prog->Done();
 	      return false;
 	    }
 
 	  if(rename(newstr.c_str(), statefile.c_str()) != 0)
 	    {
 	      _error->Errno("save_selection_list", _("couldn't replace %s with %s"), statefile.c_str(), newstr.c_str());
-	      prog.Done();
+	      if (Prog)
+		Prog->Done();
 	      return false;
 	    }
 	}
@@ -983,63 +989,80 @@ bool aptitudeDepCache::save_selection_list(OpProgress &prog,
       if (! dpkg_selections_saved)
 	{
 	  _error->Error(_("failed to save selections to dpkg database"));
-	  prog.Done();
+	  if (Prog)
+	    Prog->Done();
 	  return false;
 	}
     }
 
-  prog.Done();
+  if (Prog)
+    Prog->Done();
+
   return true;
 }
 
-void aptitudeDepCache::set_new_flag(const pkgCache::PkgIterator &pkg,
+void aptitudeDepCache::set_new_flag(const pkgCache::PkgIterator& pkg,
 				    bool is_new)
 {
-  if(read_only && !read_only_permission())
+  if (read_only && !read_only_permission())
     {
-      if(group_level == 0)
+      if (group_level == 0)
 	read_only_fail();
       return;
     }
 
-  aptitude_state &estate=get_ext_state(pkg);
+  aptitude_state& estate = get_ext_state(pkg);
 
-  if(estate.new_package && !is_new)
+  if (estate.new_package && !is_new)
     {
       --new_package_count;
-      estate.new_package=is_new;
+      estate.new_package = is_new;
     }
-  else if(!estate.new_package && is_new)
+  else if (!estate.new_package && is_new)
     {
       ++new_package_count;
-      estate.new_package=is_new;
+      estate.new_package = is_new;
     }
 }
 
 void aptitudeDepCache::forget_new(undoable **undoer)
 {
-  if(read_only && !read_only_permission())
+  std::vector<pkgCache::PkgIterator> pkg_its;
+  for (pkgCache::PkgIterator it = PkgBegin(); !it.end(); ++it)
     {
-      if(group_level == 0)
+      pkg_its.push_back(it);
+    }
+
+  forget_new(undoer, pkg_its);
+}
+
+void aptitudeDepCache::forget_new(undoable **undoer, const std::vector<pkgCache::PkgIterator>& pkg_its)
+{
+  if (read_only && !read_only_permission())
+    {
+      if (group_level == 0)
 	read_only_fail();
       return;
     }
 
-  forget_undoer *undo=undoer?new forget_undoer(this):NULL;
+  forget_undoer* undo = undoer ? new forget_undoer(this) : NULL;
 
-  for(pkgCache::PkgIterator i=PkgBegin(); !i.end(); i++)
-    if(package_states[i->ID].new_package)
-      {
-	dirty=true;
-	package_states[i->ID].new_package=false;
-	if(undo)
-	  undo->add_item(i);
-      }
+  for (pkgCache::PkgIterator i : pkg_its)
+    {
+      if (package_states[i->ID].new_package)
+	{
+	  if (package_states[i->ID].new_package && (new_package_count > 0))
+	    --new_package_count;
 
-  new_package_count=0;
+	  dirty = true;
+	  package_states[i->ID].new_package = false;
+	  if (undo)
+	    undo->add_item(i);
+	}
+    }
 
-  if(undoer && undo && !undo->empty())
-    *undoer=undo;
+  if (undoer && undo && !undo->empty())
+    *undoer = undo;
   else
     delete undo;
 
@@ -1192,6 +1215,10 @@ void aptitudeDepCache::internal_mark_install(const PkgIterator &Pkg,
   // here and restore it afterwards:
   bool previously_auto = ((*this)[Pkg].Flags & Flag::Auto) != 0;
 
+  bool final_auto = previously_auto;
+  if (set_to_manual)
+    final_auto = false;
+
   if(!ReInstall)
     {
       pkgDepCache::MarkInstall(Pkg, AutoInst);
@@ -1201,14 +1228,12 @@ void aptitudeDepCache::internal_mark_install(const PkgIterator &Pkg,
 
   pkgDepCache::SetReInstall(Pkg, ReInstall);
 
-  MarkAuto(Pkg, previously_auto);
-
-  if(set_to_manual)
-    MarkAuto(Pkg, false);
+  MarkAuto(Pkg, final_auto);
 
   get_ext_state(Pkg).selection_state=pkgCache::State::Install;
   get_ext_state(Pkg).reinstall=ReInstall;
   get_ext_state(Pkg).forbidver="";
+  get_ext_state(Pkg).previously_auto_package = final_auto;
 }
 
 void aptitudeDepCache::mark_delete(const PkgIterator &Pkg,
@@ -1216,6 +1241,15 @@ void aptitudeDepCache::mark_delete(const PkgIterator &Pkg,
 				   bool unused_delete,
 				   undo_group *undo)
 {
+  // refuse to remove itself -- see #319782, #568548
+  //
+  // but allow purge if already removed (#822331)
+  if (!Pkg.end() && is_installed(Pkg) && string("aptitude") == Pkg.Name())
+    {
+      _error->Error(_("Cannot remove aptitude within aptitude"));
+      return;
+    }
+
   if(read_only && !read_only_permission())
     {
       if(group_level == 0)
@@ -1243,6 +1277,15 @@ void aptitudeDepCache::internal_mark_delete(const PkgIterator &Pkg,
 					    bool unused_delete,
 					    std::vector<unsigned int>& unused_already_visited)
 {
+  // refuse to remove itself -- see #319782, #568548
+  //
+  // but allow purge if already removed (#822331)
+  if (!Pkg.end() && is_installed(Pkg) && string("aptitude") == Pkg.Name())
+    {
+      _error->Error(_("Cannot remove aptitude within aptitude"));
+      return;
+    }
+
   // honour ::Purge-Unused in the main entry point for removing packages, it
   // should catch cases of automatically installed and unused packages not
   // purged (#724034 and others)
@@ -1309,8 +1352,13 @@ void aptitudeDepCache::internal_mark_delete(const PkgIterator &Pkg,
   // not to purge unused lightly, can cause data loss -- see comments in #661188
   Purge = aptcfg->FindB(PACKAGE "::Purge-Unused", false);
 
-  bool follow_recommends = aptcfg->FindB("APT::Install-Recommends", true) || aptcfg->FindB(PACKAGE "::Keep-Recommends", false);
-  bool follow_suggests   = aptcfg->FindB(PACKAGE "::Keep-Suggests", false) || aptcfg->FindB(PACKAGE "::Suggests-Important", false);
+  bool keep_recommends_installed =
+    aptcfg->FindB("APT::Install-Recommends", true)
+    || aptcfg->FindB("APT::AutoRemove::RecommendsImportant", true)
+    || aptcfg->FindB(PACKAGE "::Keep-Recommends", false);
+  bool keep_suggests_installed   =
+    aptcfg->FindB("APT::AutoRemove::SuggestsImportant", true)
+    || aptcfg->FindB(PACKAGE "::Keep-Suggests", false);
 
   for (pkgCache::DepIterator dep = Pkg.CurrentVer().DependsList(); !dep.end(); ++dep)
     {
@@ -1324,8 +1372,8 @@ void aptitudeDepCache::internal_mark_delete(const PkgIterator &Pkg,
       // consider only these type of dependencies
       if (! ((dep->Type == pkgCache::Dep::Depends) ||
 	     (dep->Type == pkgCache::Dep::PreDepends) ||
-	     (dep->Type == pkgCache::Dep::Recommends && follow_recommends) ||
-	     (dep->Type == pkgCache::Dep::Suggests   && follow_suggests)))
+	     (dep->Type == pkgCache::Dep::Recommends && keep_recommends_installed) ||
+	     (dep->Type == pkgCache::Dep::Suggests   && keep_suggests_installed)))
 	{
 	  continue;
 	}
@@ -1341,13 +1389,13 @@ void aptitudeDepCache::internal_mark_delete(const PkgIterator &Pkg,
 	  for (pkgCache::PrvIterator dep_prv = dep_pkg.ProvidesList(); !dep_prv.end(); ++dep_prv)
 	    {
 	      // virtual package itself
-	      if (! can_remove_autoinstalled(dep_pkg, (*this), follow_recommends, follow_suggests)) {
+	      if (! can_remove_autoinstalled(dep_pkg, (*this), keep_recommends_installed, keep_suggests_installed)) {
 		continue;
 	      }
 
 	      // real package
 	      if (! is_auto_installed((*this)[dep_prv.OwnerPkg()]) ||
-		  ! can_remove_autoinstalled(dep_prv.OwnerPkg(), (*this), follow_recommends, follow_suggests)) {
+		  ! can_remove_autoinstalled(dep_prv.OwnerPkg(), (*this), keep_recommends_installed, keep_suggests_installed)) {
 		continue;
 	      }
 
@@ -1379,7 +1427,7 @@ void aptitudeDepCache::internal_mark_delete(const PkgIterator &Pkg,
 
       if (is_installed(dep_pkg) && is_auto && is_not_required)
 	{
-	  if (! can_remove_autoinstalled(dep_pkg, (*this), follow_recommends, follow_suggests)) {
+	  if (! can_remove_autoinstalled(dep_pkg, (*this), keep_recommends_installed, keep_suggests_installed)) {
 	    continue;
 	  }
 
@@ -1421,11 +1469,16 @@ void aptitudeDepCache::internal_mark_keep(const PkgIterator &Pkg, bool Automatic
   if(was_garbage_removed)
     MarkAuto(Pkg, false);
 
-  set_candidate_version(GetCandidateVer(Pkg), NULL);
+  set_candidate_version(GetCandidateVersion(Pkg), NULL);
 
   pkgDepCache::MarkKeep(Pkg, false, !Automatic);
   pkgDepCache::SetReInstall(Pkg, false);
   get_ext_state(Pkg).reinstall=false;
+  get_ext_state(Pkg).forbidver="";
+
+  // explicitly mark auto-installed, sometimes apt does not apply it properly in
+  // some cases -- see #508428
+  pkgDepCache::MarkAuto(Pkg, Automatic);
 
   if(Pkg.CurrentVer().end())
     {
@@ -1475,7 +1528,7 @@ void aptitudeDepCache::set_candidate_version(const VerIterator &ver,
 
       aptitude_state &estate = get_ext_state(ver.ParentPkg());
 
-      if(ver!=GetCandidateVer(ver.ParentPkg()))
+      if(ver!=GetCandidateVersion(ver.ParentPkg()))
 	estate.candver=ver.VerStr();
       else
 	estate.candver="";
@@ -1626,7 +1679,7 @@ namespace
   };
 }
 
-void aptitudeDepCache::attach_user_tag(const PkgIterator &pkg,
+bool aptitudeDepCache::attach_user_tag(const PkgIterator &pkg,
 				       const std::string &tag,
 				       undo_group *undo)
 {
@@ -1634,33 +1687,48 @@ void aptitudeDepCache::attach_user_tag(const PkgIterator &pkg,
     {
       if(group_level == 0)
 	read_only_fail();
-      return;
+      return false;
     }
 
-  // Find the tag in our cache or add it.
-  typedef std::map<std::string, user_tag_reference>::const_iterator index_ref;
-  index_ref found = user_tags_index.find(tag);
-
-  if(found == user_tags_index.end())
+  // check for valid tag
+  if ( ! user_tags.check_valid(tag) )
     {
-      user_tag_reference loc = user_tags.size();
-      user_tags.push_back(tag);
-      std::pair<index_ref, bool> tmp(user_tags_index.insert(std::make_pair(tag, loc)));
-      found = tmp.first;
+      return false;
     }
 
-  std::pair<std::set<user_tag>::const_iterator, bool> insert_result =
-    get_ext_state(pkg).user_tags.insert(user_tag(found->second));
+  // find the tag in cache or add it
+  user_tag_reference ref;
+  user_tags.add(tag, ref);
 
-  if(insert_result.second)
+  // try to add to package
+  aptitude_state& estate = get_ext_state(pkg);
+  auto found = estate.user_tags.find(user_tag{ref});
+  if (found != estate.user_tags.end())
     {
-      dirty = true;
-      if(undo != NULL)
-	undo->add_item(new attach_user_tag_undoer(this, pkg, tag));
+      // tag already present, return "OK"
+      _error->Notice(_("User-tag '%s' already present for %s"), tag.c_str(), pkg.FullName(true).c_str());
+      return true;
+    }
+  else
+    {
+      auto insert_result = estate.user_tags.insert(user_tag{ref});
+      if (insert_result.second)
+	{
+	  dirty = true;
+	  if (undo != NULL)
+	    undo->add_item(new attach_user_tag_undoer(this, pkg, tag));
+
+	  return true;
+	}
+      else
+	{
+	  _error->Error(_("Could not add user-tag '%s' to package %s"), tag.c_str(), pkg.FullName(true).c_str());
+	  return false;
+	}
     }
 }
 
-void aptitudeDepCache::detach_user_tag(const PkgIterator &pkg,
+bool aptitudeDepCache::detach_user_tag(const PkgIterator &pkg,
 				       const std::string &tag,
 				       undo_group *undo)
 {
@@ -1668,24 +1736,59 @@ void aptitudeDepCache::detach_user_tag(const PkgIterator &pkg,
     {
       if(group_level == 0)
 	read_only_fail();
-      return;
+      return false;
     }
 
-  std::map<std::string, user_tag_reference>::const_iterator found =
-    user_tags_index.find(tag);
-
-  if(found == user_tags_index.end())
-    return;
-
-  std::set<user_tag>::size_type num_erased =
-    get_ext_state(pkg).user_tags.erase(user_tag(found->second));
-
-  if(num_erased > 0)
+  // check for valid tag
+  if ( ! user_tags.check_valid(tag) )
     {
-      dirty = true;
-      if(undo != NULL)
-	undo->add_item(new detach_user_tag_undoer(this, pkg, tag));
+      return false;
     }
+
+  user_tag_reference tag_ref = user_tags.get_ref(tag);
+  if (tag_ref < 0)
+    {
+      _error->Error(_("Could not find valid user-tag '%s'"), tag.c_str());
+      return false;
+    }
+  else
+    {
+      std::set<user_tag>::size_type num_erased = get_ext_state(pkg).user_tags.erase(user_tag{tag_ref});
+      if (num_erased > 0)
+	{
+	  dirty = true;
+	  if(undo != NULL)
+	    undo->add_item(new detach_user_tag_undoer(this, pkg, tag));
+
+	  return true;
+	}
+      else
+	{
+	  _error->Error(_("Could not remove user-tag '%s' from package %s"), tag.c_str(), pkg.FullName(true).c_str());
+	  return false;
+	}
+    }
+}
+
+std::vector<std::string> aptitudeDepCache::get_user_tags(const PkgIterator& pkg)
+{
+  // sanity check
+  if (pkg.end())
+    return {};
+
+  auto estate = get_ext_state(pkg);
+
+  std::vector<std::string> all_tags;
+  all_tags.reserve(estate.user_tags.size());
+
+  for (const auto& it : estate.user_tags)
+    {
+      all_tags.push_back(user_tags.deref_user_tag(it));
+    }
+
+  std::sort(all_tags.begin(), all_tags.end());
+
+  return all_tags;
 }
 
 bool aptitudeDepCache::all_upgrade(bool with_autoinst, undo_group *undo)
@@ -2324,7 +2427,10 @@ void aptitudeDepCache::apply_solution(const generic_solution<aptitude_universe>
 	{
 	  aptitude_resolver_version ver(i->get_ver());
 	  LOG_TRACE(logger, "Adding version chosen by the resolver: " << ver);
-	  versions.push_back(std::make_pair(ver, true));
+	  // auto_installed true by default because "install_version" is very
+	  // broad, but might mean to "keep the same version"
+	  constexpr bool auto_installed = true;
+	  versions.push_back(std::make_pair(ver, auto_installed));
 	}
       else
 	LOG_TRACE(logger, "Skipping " << *i << ": it is not a version install.");
@@ -2359,7 +2465,11 @@ void aptitudeDepCache::apply_solution(const generic_solution<aptitude_universe>
 		    << " at its current version ("
 		    << curver.VerStr() << ")");
 
-	  internal_mark_keep(pkg, is_auto, false);
+	  // auto_installed true because "install_version" is very broad, but
+	  // might mean to "keep the same version"
+	  bool was_auto = is_auto_installed(pkg);
+
+	  internal_mark_keep(pkg, was_auto, false);
 	}
       else
 	// install a particular version that's not the current one.
@@ -2395,8 +2505,11 @@ aptitudeCacheFile::~aptitudeCacheFile()
   delete Policy;
 }
 
-bool aptitudeCacheFile::Open(OpProgress &Progress, bool do_initselections,
-			     bool WithLock, const char *status_fname)
+bool aptitudeCacheFile::Open(OpProgress* Progress,
+			     bool do_initselections,
+			     bool WithLock,
+			     const char* status_fname,
+			     bool reset_reinstall)
 {
   if(WithLock)
     {
@@ -2414,8 +2527,9 @@ bool aptitudeCacheFile::Open(OpProgress &Progress, bool do_initselections,
     return _error->Error(_("The list of sources could not be read."));
 
   // Read the caches:
-  bool Res=pkgMakeStatusCache(List, Progress, &Map, !WithLock);
-  Progress.Done();
+  bool Res = pkgCacheGenerator::MakeStatusCache(List, Progress, &Map, !WithLock);
+  if (Progress)
+    Progress->Done();
 
   if(!Res)
     return _error->Error(_("The package lists or status file could not be parsed or opened."));
@@ -2433,15 +2547,24 @@ bool aptitudeCacheFile::Open(OpProgress &Progress, bool do_initselections,
   if(ReadPinFile(*Policy) == false || ReadPinDir(*Policy) == false)
     return false;
 
-  DCache=new aptitudeDepCache(Cache, Policy);
-  if(_error->PendingError())
-    return false;
+  {
+    DCache=new aptitudeDepCache(Cache, Policy);
+    if (_error->PendingError())
+      {
+	_error->Error(_("Could not create dependency cache"));
+	return false;
+      }
 
-  DCache->Init(&Progress, WithLock, do_initselections, status_fname);
-  Progress.Done();
+    bool init_result = DCache->Init(Progress, WithLock, do_initselections, status_fname, reset_reinstall);
+    if (Progress)
+      Progress->Done();
 
-  if(_error->PendingError())
-    return false;
+    if (!init_result || _error->PendingError())
+      {
+	_error->Error(_("Could not initialize dependency cache"));
+	return false;
+      }
+  }
 
   return true;
 }
