@@ -2,6 +2,7 @@
 //
 //   Copyright (C) 2005, 2007-2008, 2010 Daniel Burrows
 //   Copyright (C) 2014 Daniel Hartwig
+//   Copyright (C) 2016 Manuel A. Fernandez Montecelo
 //
 //   This program is free software; you can redistribute it and/or
 //   modify it under the terms of the GNU General Public License as
@@ -15,8 +16,8 @@
 //
 //   You should have received a copy of the GNU General Public License
 //   along with this program; see the file COPYING.  If not, write to
-//   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-//   Boston, MA 02111-1307, USA.
+//   the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
+//   Boston, MA 02110-1301, USA.
 
 #include "tags.h"
 
@@ -204,19 +205,28 @@ static bool load_tags_from_debtags(OpProgress *progress)
       return false;
     }
 
-  const unsigned long long m = F.Size();
+  const unsigned long long file_size = F.Size();
 
-  if(progress != NULL)
-    progress->OverallProgress(0, m, 1,
+  // last percent shown in progress -- do not update on every cycle
+  int last_pct_shown = 0;
+  if (progress)
+    progress->OverallProgress(0, file_size, 1,
                               _("Building tag database"));
 
   const unsigned long long buf_size = 4096;
   char buf[buf_size];
   while(F.ReadLine(buf, buf_size) != NULL)
     {
-      if(progress != NULL)
-        progress->OverallProgress(F.Tell(), m, 1,
-                                  _("Building tag database"));
+      if (progress)
+	{
+	  int pct = (100*F.Tell()) / file_size;
+	  if ((pct % 10 == 1) && last_pct_shown != pct)
+	    {
+	      last_pct_shown = pct;
+	      progress->OverallProgress(F.Tell(), file_size, 1,
+					_("Building tag database"));
+	    }
+	}
 
       const char *sep = strstr(buf, ": ");
       if(sep == NULL)
@@ -241,8 +251,12 @@ static bool load_tags_from_debtags(OpProgress *progress)
         tags->insert(*t);
     }
 
-  if(progress != NULL)
-    progress->Done();
+  if (progress)
+    {
+      progress->OverallProgress(file_size, file_size, 1,
+				_("Building tag database"));
+      progress->Done();
+    }
 
   return true;
 }
@@ -256,23 +270,40 @@ static bool load_tags_from_verfiles(OpProgress *progress)
     for(pkgCache::VerIterator v = p.VersionList(); !v.end(); ++v)
       for(pkgCache::VerFileIterator vf = v.FileList();
 	  !vf.end(); ++vf)
-	verfiles.push_back(loc_pair(v, vf));
+	  verfiles.push_back(loc_pair(v, vf));
 
   sort(verfiles.begin(), verfiles.end(), location_compare());
 
-  progress->OverallProgress(0, verfiles.size(), 1,
-                            _("Building tag database"));
-  size_t n=0;
+  size_t progress_num = 0;
+  int progress_total = verfiles.size();
+  // only update if we're going to increase 10% or so, minimum 1 (to avoid
+  // divide by zero)
+  int update_progress_10pct = std::max(progress_total / 10, 1);
+  if (progress)
+    progress->OverallProgress(0, progress_total, 1, _("Building tag database"));
+
   for(std::vector<loc_pair>::iterator i=verfiles.begin();
       i!=verfiles.end(); ++i)
     {
       insert_tags(i->first, i->second);
-      ++n;
-      progress->OverallProgress(n, verfiles.size(), 1,
-                                _("Building tag database"));
+
+      if (progress)
+	{
+	  // don't update on every cycle
+	  if ((++progress_num % update_progress_10pct) == 1)
+	    {
+	      progress->OverallProgress(progress_num, progress_total, 1,
+					_("Building tag database"));
+	    }
+	}
     }
 
-  progress->Done();
+  if (progress)
+    {
+      progress->OverallProgress(progress_total, progress_total, 1,
+				_("Building tag database"));
+      progress->Done();
+    }
 
   return true;
 }
