@@ -1,7 +1,7 @@
 // main.cc  (neé testscr.cc)
 //
 //  Copyright 1999-2011 Daniel Burrows
-//  Copyright 2014-2015 Manuel A. Fernandez Montecelo
+//  Copyright 2014-2016 Manuel A. Fernandez Montecelo
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -15,8 +15,8 @@
 //
 //  You should have received a copy of the GNU General Public License
 //  along with this program; see the file COPYING.  If not, write to
-//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-//  Boston, MA 02111-1307, USA.
+//  the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
+//  Boston, MA 02110-1301, USA.
 //
 //  Tests the various screen-output mechanisms
 
@@ -58,15 +58,16 @@
 #include <cwidget/toplevel.h>
 #include <cwidget/dialogs.h>
 
+#include <cmdline/cmdline_apt_proxy.h>
 #include <cmdline/cmdline_changelog.h>
 #include <cmdline/cmdline_check_resolver.h>
 #include <cmdline/cmdline_clean.h>
 #include <cmdline/cmdline_common.h>
 #include <cmdline/cmdline_do_action.h>
-#include <cmdline/cmdline_download.h>
 #include <cmdline/cmdline_dump_resolver.h>
 #include <cmdline/cmdline_extract_cache_subset.h>
 #include <cmdline/cmdline_forget_new.h>
+#include <cmdline/cmdline_mark.h>
 #include <cmdline/cmdline_moo.h>
 #include <cmdline/cmdline_prompt.h>
 #include <cmdline/cmdline_search.h>
@@ -94,6 +95,7 @@
 #endif
 
 #include <fstream>
+#include <locale>
 
 #include "loggers.h"
 #include "progress.h"
@@ -148,8 +150,9 @@ std::string handle_wtomb_error(int error,
 
 static void show_version()
 {
-  printf(_("%s %s compiled at %s %s\n"),
-	   PACKAGE, VERSION, __DATE__, __TIME__);
+  // don't use __DATE__ or __TIME__, for reproducible builds
+  printf("%s %s\n", PACKAGE, VERSION);
+
 #ifdef __GNUC__
   printf(_("Compiler: g++ %s\n"), __VERSION__);
 #endif
@@ -189,69 +192,74 @@ static void usage()
   printf("\n");
   printf(_("       aptitude [options] <action> ..."));
   printf("\n");
-  printf(_("  Actions (if none is specified, aptitude will enter interactive mode):\n\n"));
-  printf(_(" install      - Install/upgrade packages.\n"));
-  printf(_(" remove       - Remove packages.\n"));
-  printf(_(" purge        - Remove packages and their configuration files.\n"));
-  printf(_(" hold         - Place packages on hold.\n"));
-  printf(_(" unhold       - Cancel a hold command for a package.\n"));
-  printf(_(" markauto     - Mark packages as having been automatically installed.\n"));
-  printf(_(" unmarkauto   - Mark packages as having been manually installed.\n"));
-  printf(_(" forbid-version - Forbid aptitude from upgrading to a specific package version.\n"));
-  printf(_(" update       - Download lists of new/upgradable packages.\n"));
-  printf(_(" safe-upgrade - Perform a safe upgrade.\n"));
-  printf(_(" full-upgrade - Perform an upgrade, possibly installing and removing packages.\n"));
-  printf(_(" build-dep    - Install the build-dependencies of packages.\n"));
-  printf(_(" forget-new   - Forget what packages are \"new\".\n"));
-  printf(_(" search       - Search for a package by name and/or expression.\n"));
-  printf(_(" show         - Display detailed information about a package.\n"));
-  printf(_(" versions     - Displays the versions of specified packages.\n"));
-  printf(_(" clean        - Erase downloaded package files.\n"));
-  printf(_(" autoclean    - Erase old downloaded package files.\n"));
-  printf(_(" changelog    - View a package's changelog.\n"));
-  printf(_(" download     - Download the .deb file for a package.\n"));
-  printf(_(" reinstall    - Download and (possibly) reinstall a currently installed package.\n"));
-  printf(_(" why          - Show the manually installed packages that require a package, or\n"
-           "                why one or more packages would require the given package.\n"));
-  printf(_(" why-not      - Show the manually installed packages that lead to a conflict\n"
-           "                with the given package, or why one or more packages would\n"
-           "                lead to a conflict with the given package if installed.\n"));
   printf("\n");
-  printf(_("  Options:\n"));
-  printf(_(" -h             This help text.\n"));
+  printf(_("Actions (if none is specified, aptitude will enter interactive mode):\n\n"));
+  printf(_(" install         Install/upgrade packages.\n"));
+  printf(_(" remove          Remove packages.\n"));
+  printf(_(" purge           Remove packages and their configuration files.\n"));
+  printf(_(" hold            Place packages on hold.\n"));
+  printf(_(" unhold          Cancel a hold command for a package.\n"));
+  printf(_(" markauto        Mark packages as having been automatically installed.\n"));
+  printf(_(" unmarkauto      Mark packages as having been manually installed.\n"));
+  printf(_(" forbid-version  Forbid aptitude from upgrading to a specific package version.\n"));
+  printf(_(" update          Download lists of new/upgradable packages.\n"));
+  printf(_(" safe-upgrade    Perform a safe upgrade.\n"));
+  printf(_(" full-upgrade    Perform an upgrade, possibly installing and removing packages.\n"));
+  printf(_(" build-dep       Install the build-dependencies of packages.\n"));
+  printf(_(" forget-new      Forget what packages are \"new\".\n"));
+  printf(_(" search          Search for a package by name and/or expression.\n"));
+  printf(_(" show            Display detailed info about a package.\n"));
+  printf(_(" showsrc         Display detailed info about a source package (apt wrapper).\n"));
+  printf(_(" versions        Displays the versions of specified packages.\n"));
+  printf(_(" clean           Erase downloaded package files.\n"));
+  printf(_(" autoclean       Erase old downloaded package files.\n"));
+  printf(_(" changelog       View a package's changelog.\n"));
+  printf(_(" download        Download the .deb file for a package (apt wrapper).\n"));
+  printf(_(" source          Download source package (apt wrapper).\n"));
+  printf(_(" reinstall       Reinstall a currently installed package.\n"));
+  printf(_(" why             Explain why a particular package should be installed.\n"));
+  printf(_(" why-not         Explain why a particular package cannot be installed.\n"));
+  printf("\n");
+  printf(_(" add-user-tag    Add user tag to packages/patterns.\n"));
+  printf(_(" remove-user-tag Remove user tag from packages/patterns.\n"));
+  printf("\n");
+  printf(_("Options:\n"));
+  printf(_(" -h              This help text.\n"));
 #ifdef HAVE_GTK
-  printf(_(" --gui          Use the GTK GUI even if disabled in the configuration.\n"));
+  printf(_(" --gui           Use the GTK GUI even if disabled in the configuration.\n"));
 #endif
-  printf(_(" --no-gui       Do not use the GTK GUI even if available.\n"));
+  printf(_(" --no-gui        Do not use the GTK GUI even if available.\n"));
 #ifdef HAVE_QT
-  printf(_(" --qt           Use the Qt GUI.\n"));
-  printf(_(" --no-qt        Do not use the Qt GUI even if enabled in the configuration.\n"));
+  printf(_(" --qt            Use the Qt GUI.\n"));
+  printf(_(" --no-qt         Do not use the Qt GUI even if enabled in the configuration.\n"));
 #endif
-  printf(_(" -s             Simulate actions, but do not actually perform them.\n"));
-  printf(_(" -d             Only download packages, do not install or remove anything.\n"));
-  printf(_(" -P             Always prompt for confirmation of actions.\n"));
-  printf(_(" -y             Assume that the answer to simple yes/no questions is 'yes'.\n"));
-  printf(_(" -F format      Specify a format for displaying search results; see the manual.\n"));
-  printf(_(" -O order       Specify how search results should be sorted; see the manual.\n"));
-  printf(_(" -w width       Specify the display width for formatting search results.\n"));
-  printf(_(" -f             Aggressively try to fix broken packages.\n"));
-  printf(_(" -V             Show which versions of packages are to be installed.\n"));
-  printf(_(" -D             Show the dependencies of automatically changed packages.\n"));
-  printf(_(" -Z             Show the change in installed size of each package.\n"));
-  printf(_(" -v             Display extra information. (may be supplied multiple times).\n"));
-  printf(_(" -t [release]   Set the release from which packages should be installed.\n"));
-  printf(_(" -q             In command-line mode, suppress the incremental progress\n"
-           "                indicators.\n"));
-  printf(_(" -o key=val     Directly set the configuration option named 'key'.\n"));
-  printf(_(" --with(out)-recommends	Specify whether or not to treat recommends as\n"
-           "                strong dependencies.\n"));
-  printf(_(" -S fname       Read the aptitude extended status info from fname.\n"));
-  printf(_(" -u             Download new package lists on startup.\n"));
+  printf(_(" -s              Simulate actions, but do not actually perform them.\n"));
+  printf(_(" -d              Only download packages, do not install or remove anything.\n"));
+  printf(_(" -P              Always prompt for confirmation of actions.\n"));
+  printf(_(" -y              Assume that the answer to simple yes/no questions is 'yes'.\n"));
+  printf(_(" -F format       Specify a format for displaying search results; see the manual.\n"));
+  printf(_(" -O order        Specify how search results should be sorted; see the manual.\n"));
+  printf(_(" -w width        Specify the display width for formatting search results.\n"));
+  printf(_(" -f              Aggressively try to fix broken packages.\n"));
+  printf(_(" -V              Show which versions of packages are to be installed.\n"));
+  printf(_(" -D              Show the dependencies of automatically changed packages.\n"));
+  printf(_(" -Z              Show the change in installed size of each package.\n"));
+  printf(_(" -v              Display extra information. (may be supplied multiple times).\n"));
+  printf(_(" -t [release]    Set the release from which packages should be installed.\n"));
+  printf(_(" -q              In command-line mode, suppress the incremental progress\n"
+           "                  indicators.\n"));
+  printf(_(" -o key=val      Directly set the configuration option named 'key'.\n"));
+  printf(_(" --with(out)-recommends     Specify whether or not to treat recommends as\n"
+           "                            strong dependencies.\n"));
+  printf(_(" -S fname        Read the aptitude extended status info from fname.\n"));
+  printf(_(" -u              Download new package lists on startup.\n"));
   printf(_("                  (terminal interface only)\n"));
-  printf(_(" -i             Perform an install run on startup.\n"));
+  printf(_(" -i              Perform an install run on startup.\n"));
   printf(_("                  (terminal interface only)\n"));
   printf("\n");
-  printf(_("                  This aptitude does not have Super Cow Powers.\n"));
+  printf(_("See the manual page for a complete list and description of all the options.\n"));
+  printf("\n");
+  printf(_("This aptitude does not have Super Cow Powers.\n"));
 }
 
 // This handles options with no single-character equivalent
@@ -373,27 +381,15 @@ namespace
     {
       using namespace logging;
 
-      // TRANSLATORS: This is a log level that the user can pass on
-      // the command-line or set in the configuration file.
-      add_level(N_("trace"), TRACE_LEVEL);
-      // TRANSLATORS: This is a log level that the user can pass on
-      // the command-line or set in the configuration file.
-      add_level(N_("debug"), DEBUG_LEVEL);
-      // TRANSLATORS: This is a log level that the user can pass on
-      // the command-line or set in the configuration file.
-      add_level(N_("info"), INFO_LEVEL);
-      // TRANSLATORS: This is a log level that the user can pass on
-      // the command-line or set in the configuration file.
-      add_level(N_("warn"), WARN_LEVEL);
-      // TRANSLATORS: This is a log level that the user can pass on
-      // the command-line or set in the configuration file.
-      add_level(N_("error"), ERROR_LEVEL);
-      // TRANSLATORS: This is a log level that the user can pass on
-      // the command-line or set in the configuration file.
-      add_level(N_("fatal"), FATAL_LEVEL);
-      // TRANSLATORS: This is a log level that the user can pass on
-      // the command-line or set in the configuration file.
-      add_level(N_("off"), OFF_LEVEL);
+      // TRANSLATORS: These are not translatable anymore
+
+      add_level("trace", TRACE_LEVEL);
+      add_level("debug", DEBUG_LEVEL);
+      add_level("info",  INFO_LEVEL);
+      add_level("warn",  WARN_LEVEL);
+      add_level("error", ERROR_LEVEL);
+      add_level("fatal", FATAL_LEVEL);
+      add_level("off",   OFF_LEVEL);
 
       std::vector<std::pair<std::string, log_level> >
 	tmp(levels.begin(), levels.end());
@@ -459,8 +455,8 @@ namespace
 
     if(!level)
       {
-	// TRANSLATORS: both the translated and the untranslated
-	// log level names are accepted here.
+	// TRANSLATORS: levels are not translatable anymore
+
 	_error->Error(_("Unknown log level name \"%s\" (expected \"trace\", \"debug\", \"info\", \"warn\", \"error\", \"fatal\", or \"off\")."),
 		      level_name.c_str());
 	return;
@@ -666,7 +662,16 @@ int main(int argc, char *argv[])
   //cw::util::transcode_mbtow_err=handle_mbtow_error;
   //cw::util::transcode_wtomb_err=handle_wtomb_error;
 
-  setlocale(LC_ALL, "");
+  // Use the C++ classes if possible, otherwise if ::global() is not called the
+  // internal state of std::locale is not initialized properly
+  //
+  // it can throw an exception if the locale defined in the environment is not
+  // valid
+  try {
+    std::locale::global(std::locale(""));
+  } catch (...) {
+    setlocale(LC_ALL, "");
+  }
   bindtextdomain(PACKAGE, LOCALEDIR);
   textdomain(PACKAGE);
 
@@ -909,7 +914,15 @@ int main(int argc, char *argv[])
 	      queue_only=true;
 	      break;
 	    case OPTION_PURGE_UNUSED:
-	      aptcfg->Set(PACKAGE "::Purge-Unused", "true");
+	      if (string("--purge-unused") == argv[optind-1])
+		{
+		  aptcfg->Set(PACKAGE "::Purge-Unused", "true");
+		}
+	      else
+		{
+		  fprintf(stderr, _("Invalid option \"%s\" (only \"--purge-unused\" is valid).  Perhaps you want the \"purge\" action?\n"), argv[optind-1]);
+		  exit(EXIT_FAILURE);
+		}
 	      break;
 	    case OPTION_ADD_USER_TAG:
 	    case OPTION_REMOVE_USER_TAG:
@@ -1065,21 +1078,20 @@ int main(int argc, char *argv[])
     }
 
   aptitude::why::roots_string_mode why_display_mode;
-  if(show_why_summary_mode == "no-summary" || show_why_summary_mode == _("no-summary"))
+  if(show_why_summary_mode == "no-summary")
     why_display_mode = aptitude::why::no_summary;
-  else if(show_why_summary_mode == "first-package" || show_why_summary_mode == _("first-package"))
+  else if(show_why_summary_mode == "first-package")
     why_display_mode = aptitude::why::show_requiring_packages;
-  else if(show_why_summary_mode == "first-package-and-type" || show_why_summary_mode == _("first-package-and-type"))
+  else if(show_why_summary_mode == "first-package-and-type")
     why_display_mode = aptitude::why::show_requiring_packages_and_strength;
-  else if(show_why_summary_mode == "all-packages" || show_why_summary_mode == _("all-packages"))
+  else if(show_why_summary_mode == "all-packages")
     why_display_mode = aptitude::why::show_chain;
-  else if(show_why_summary_mode == "all-packages-with-dep-versions" || show_why_summary_mode == _("all-packages-with-dep-versions"))
+  else if(show_why_summary_mode == "all-packages-with-dep-versions")
     why_display_mode = aptitude::why::show_chain_with_versions;
   else
     {
-      // TRANSLATORS: "why" here is the aptitude command name and should not
-      // be translated.  Both the translated and the untranslated log level
-      // names are accepted here.
+      // TRANSLATORS: "why" and "modes" here are the aptitude command name and
+      // parameters, and should not be translated.
       _error->Error(_("Invalid \"why\" summary mode \"%s\": expected "
 		      "\"no-summary\", \"first-package\", \"first-package-and-type\", \"all-packages\", or \"all-packages-with-dep-versions\"."),
 		    show_why_summary_mode.c_str());
@@ -1188,6 +1200,41 @@ int main(int argc, char *argv[])
 	    return cmdline_why(argc - optind, argv + optind,
 			       status_fname, verbose,
 			       why_display_mode, true);
+	  else if ( (!strcasecmp(argv[optind], "hold")) ||
+		    (!strcasecmp(argv[optind], "unhold")) ||
+		    (!strcasecmp(argv[optind], "forbid-version")) ||
+		    (!strcasecmp(argv[optind], "markauto")) ||
+		    (!strcasecmp(argv[optind], "unmarkauto")) ||
+		    (!strcasecmp(argv[optind], "keep")) ||
+		    (!strcasecmp(argv[optind], "keep-all")) )
+	    {
+	      // require at least one argument in addition to the command
+	      int num_args = argc-optind;
+	      int error_status = 1;
+	      if ( (!strcasecmp(argv[optind], "hold")) ||
+		   (!strcasecmp(argv[optind], "unhold")) ||
+		   (!strcasecmp(argv[optind], "markauto")) ||
+		   (!strcasecmp(argv[optind], "unmarkauto")) ||
+		   (!strcasecmp(argv[optind], "forbid-version")) ||
+		   (!strcasecmp(argv[optind], "keep")) )
+		{
+		  if (num_args < 2)
+		    {
+		      fprintf(stderr, _("Command \"%s\" needs arguments\n"), argv[optind]);
+		      return error_status;
+		    }
+		}
+	      else if (!strcasecmp(argv[optind], "keep-all"))
+		{
+		  if (num_args != 1)
+		    {
+		      fprintf(stderr, _("Command \"%s\" does not accept arguments\n"), argv[optind]);
+		      return error_status;
+		    }
+		}
+
+	      return cmdline_mark(argc-optind, argv+optind, status_fname, simulate, verbose);
+	    }
 	  else if( (!strcasecmp(argv[optind], "install")) ||
 		   (!strcasecmp(argv[optind], "reinstall")) ||
 		   (!strcasecmp(argv[optind], "dist-upgrade")) ||
@@ -1196,24 +1243,11 @@ int main(int argc, char *argv[])
 		   (!strcasecmp(argv[optind], "upgrade")) ||
 		   (!strcasecmp(argv[optind], "remove")) ||
 		   (!strcasecmp(argv[optind], "purge")) ||
-		   (!strcasecmp(argv[optind], "hold")) ||
-		   (!strcasecmp(argv[optind], "unhold")) ||
-		   (!strcasecmp(argv[optind], "markauto")) ||
-		   (!strcasecmp(argv[optind], "unmarkauto")) ||
-		   (!strcasecmp(argv[optind], "forbid-version")) ||
-		   (!strcasecmp(argv[optind], "keep")) ||
-		   (!strcasecmp(argv[optind], "keep-all")) ||
 		   (!strcasecmp(argv[optind], "build-dep")) ||
 		   (!strcasecmp(argv[optind], "build-depends")))
 	    {
 	      // require at least one argument in addition to the command
-	      if ( (!strcasecmp(argv[optind], "hold")) ||
-		   (!strcasecmp(argv[optind], "unhold")) ||
-		   (!strcasecmp(argv[optind], "markauto")) ||
-		   (!strcasecmp(argv[optind], "unmarkauto")) ||
-		   (!strcasecmp(argv[optind], "forbid-version")) ||
-		   (!strcasecmp(argv[optind], "keep")) ||
-		   (!strcasecmp(argv[optind], "build-dep")) ||
+	      if ( (!strcasecmp(argv[optind], "build-dep")) ||
 		   (!strcasecmp(argv[optind], "build-depends")))
 		{
 		  int num_args = argc-optind;
@@ -1225,6 +1259,13 @@ int main(int argc, char *argv[])
 		    }
 		}
 
+	      // prevent visual preview if dumb term -- see #317928, #817276
+	      if (visual_preview && aptitude::util::is_dumb_terminal())
+		{
+		  aptitude::util::print_ncurses_dumb_terminal();
+		  exit(EXIT_FAILURE);
+		}
+
 	      return cmdline_do_action(argc-optind, argv+optind,
 				       status_fname,
 				       simulate, assume_yes, download_only,
@@ -1243,8 +1284,6 @@ int main(int argc, char *argv[])
 	  else if(!strcasecmp(argv[optind], "extract-cache-subset"))
 	    return aptitude::cmdline::extract_cache_subset(argc - optind,
 							   argv + optind);
-	  else if(!strcasecmp(argv[optind], "download"))
-	    return cmdline_download(argc-optind, argv+optind);
 	  else if(!strcasecmp(argv[optind], "changelog"))
 	    return cmdline_changelog(argc-optind, argv+optind);
 	  else if(!strcasecmp(argv[optind], "moo"))
@@ -1265,16 +1304,24 @@ int main(int argc, char *argv[])
 	    {
 	      OpTextProgress p(aptcfg->FindI("Quiet", 0));
 	      _error->DumpErrors();
-	      apt_init(&p, true);
+	      bool operation_needs_lock = false;
+	      apt_init(&p, true, operation_needs_lock, nullptr);
 	      exit(0);
 	    }
 	  else if(!strcasecmp(argv[optind], "nop-noselections"))
 	    {
 	      OpTextProgress p(aptcfg->FindI("Quiet", 0));
 	      _error->DumpErrors();
-	      apt_init(&p, false);
+	      bool operation_needs_lock = false;
+	      apt_init(&p, false, operation_needs_lock, nullptr);
 	      exit(0);
 	    }
+	  else if (!strcasecmp(argv[optind], "download") ||
+		   !strcasecmp(argv[optind], "showsrc") ||
+		   !strcasecmp(argv[optind], "source"))
+	    {
+	      return cmdline_apt_proxy(argc, argv);
+	    }
 	  else
 	    {
 	      fprintf(stderr, _("Unknown command \"%s\"\n"), argv[optind]);
@@ -1326,6 +1373,12 @@ int main(int argc, char *argv[])
     }
 #endif
 
+  if (aptitude::util::is_dumb_terminal())
+    {
+      aptitude::util::print_ncurses_dumb_terminal();
+      exit(EXIT_FAILURE);
+    }
+  else
     {
       ui_init();
 
@@ -1337,8 +1390,11 @@ int main(int argc, char *argv[])
           // reloaded anyway).  Obviously we still need them for installs,
           // since we have to get information about what to install from
           // somewhere...
-          if(!update_only)
-            apt_init(p->get_progress().unsafe_get_ref(), true, status_fname);
+          if (!update_only)
+	    {
+	      bool operation_needs_lock = true;
+	      apt_init(p->get_progress().unsafe_get_ref(), true, operation_needs_lock, status_fname);
+	    }
           if(status_fname)
             free(status_fname);
           check_apt_errors();
