@@ -1,7 +1,7 @@
 // pkg_item.cc
 //
 // Copyright 1999-2005, 2007-2009 Daniel Burrows
-// Copyright 2014-2015 Manuel A. Fernandez Montecelo
+// Copyright 2014-2016 Manuel A. Fernandez Montecelo
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -15,8 +15,8 @@
 //
 //  You should have received a copy of the GNU General Public License
 //  along with this program; see the file COPYING.  If not, write to
-//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-//  Boston, MA 02111-1307, USA.
+//  the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
+//  Boston, MA 02110-1301, USA.
 
 #include "aptitude.h"
 
@@ -24,6 +24,7 @@
 #include <cwidget/config/keybindings.h>
 #include <cwidget/dialogs.h>
 #include <cwidget/fragment.h>
+#include <cwidget/generic/util/ssprintf.h>
 #include <cwidget/generic/util/transcode.h>
 #include <cwidget/toplevel.h>
 
@@ -123,7 +124,39 @@ void pkg_item::do_highlighted_changed(bool highlighted)
 
 void pkg_item::do_select(undo_group *undo)
 {
-	(*apt_cache_file)->mark_install(package, aptcfg->FindB(PACKAGE "::Auto-Install", true), false, undo);
+  pkgCache::PkgIterator& package_to_install = package;
+
+  if (is_virtual(package))
+    {
+      std::string pkgtext;
+      std::vector<std::string> pkgnames;
+      for (pkgCache::PrvIterator prv = package.ProvidesList(); !prv.end(); ++prv)
+	{
+	  string name = prv.OwnerPkg().FullName(true);
+	  pkgnames.push_back(name);
+
+	  pkgtext += cwidget::util::ssprintf("* %s\n", name.c_str());
+	}
+
+      // messages copied from from cmdline_action.cc
+      if (pkgnames.empty())
+	{
+	  _error->Error(_("\"%s\" exists in the package database, but it is not a real package and no package provides it\n"),
+			package.FullName(true).c_str());
+	}
+      else if (pkgnames.size() == 1)
+	{
+	  // only one, install this one
+	  package_to_install = package.ProvidesList().OwnerPkg();
+	}
+      else
+	{
+	  _error->Error(_("\"%s\" is a virtual package provided by:\n\n%s\nYou must choose one to install.\n"),
+			package.FullName(true).c_str(), pkgtext.c_str());
+	}
+    }
+
+  (*apt_cache_file)->mark_install(package_to_install, aptcfg->FindB(PACKAGE "::Auto-Install", true), false, undo);
 }
 
 void pkg_item::select(undo_group *undo)
@@ -159,9 +192,11 @@ void pkg_item::hold(undo_group *undo)
 
 void pkg_item::keep(undo_group *undo)
 {
+  // copied to pkg_ver_item::keep()
+
   // Keep, don't hold, the package.
   (*apt_cache_file)->mark_keep(package,
-			       false,
+			       is_auto_installed(package),
 			       false,
 			       undo);
 }
@@ -349,71 +384,21 @@ bool pkg_item::dispatch_key(const cw::config::key &k, cw::tree *owner)
       else
 	delete grp;
     }
-  else if(bindings->key_matches(k, "BugReport"))
+  else if (bindings->key_matches(k, "CollapseTree"))
     {
-      // Try to report a bug on the package.  (ew quoting ew)
-      string cmd=string("reportbug '")+package.Name()+"'";
-
-      // Default to reporting a bug on the current version.
-      pkgCache::VerIterator ver=package.CurrentVer();
-      if(ver.end())
-	ver=visible_version();
-      if(ver.end())
-	ver=package.VersionList();
-
-      if(!ver.end())
-	cmd+=string(" -V '")+ver.VerStr()+"'";
-
-      cw::toplevel::suspend();
-
-      printf(_("Reporting a bug in %s:\n"), package.Name());
-
-      if(system(cmd.c_str()) != 0) { /* ignore */ }
-
-      cw::toplevel::resume();
-    }
-  else if(bindings->key_matches(k, "DpkgReconfigure"))
-    // Don't bother with my internal su-to-root stuff here, since I don't
-    // need to touch the package lists in the subprocess.
-    {
-      // Try to do *something*.
-      const char *sucmd=NULL;
-
-      if(getuid()==0)
-	sucmd="dpkg-reconfigure '%s'";
-      else if(access("/usr/sbin/su-to-root", X_OK)==0)
-	sucmd="/usr/sbin/su-to-root -c \"/usr/sbin/dpkg-reconfigure '%s'\"";
-      else if(access("/bin/su", X_OK)==0)
-	sucmd="/bin/su -c \"/usr/sbin/dpkg-reconfigure '%s'\"";
-      else
-	popup_widget(cw::dialogs::ok(cw::text_fragment(_("You are not root and I cannot find any way to become root.  To reconfigure this package, install the menu package, the login package, or run aptitude as root."))));
-
-      if(sucmd)
-	{
-	  cw::toplevel::suspend();
-
-	  apt_cache_file->ReleaseLock();
-
-	  printf(_("Reconfiguring %s\n"), package.FullName(true).c_str());
-
-	  char buf[512];
-	  if(sucmd)
-	    {
-              const string name(aptitude::apt::dpkg_package_name(package));
-              snprintf(buf, 512, sucmd, name.c_str());
-
-	      if(system(buf) != 0) { /* FIXME: ignore? */ }
-
-	      cout << _("Press Return to continue.") << endl;
-	      getchar();
-
-	      cw::toplevel::resume();
-	    }
-
-	  progress_ref p = gen_progress_bar();
-	  apt_reload_cache(p->get_progress().unsafe_get_ref(), true);
-	  p->destroy();
-	}
+      // hack to convert a Left key stroke to moving to the parent (several
+      // requests for this, like #241945 and #415449; with suggestions that it
+      // could be done by moving to Parent --doing this at the moment-- or
+      // folding the Parent subtree directly).
+      //
+      // unfortunately this cannot be done in a cleaner way than injecting from
+      // this tree item to the owner, setting "Left" as a keybinding for
+      // "Parent" is interpreted in other elements as well, thus disrupting
+      // unrelated things, because it's not specific to tree items.
+      //
+      // NOTE: get("") needs to be passed as uppercase, see #818046
+      cw::config::keybinding keybindings_Parent = cw::config::global_bindings.get("PARENT");
+      return ((keybindings_Parent.size() > 0) && owner && owner->dispatch_key(keybindings_Parent[0]));
     }
   else
     return pkg_tree_node::dispatch_key(k, owner);
