@@ -1,7 +1,7 @@
 // pkg_ver_item.cc
 //
 //  Copyright 1999-2005, 2007-2008, 2010 Daniel Burrows
-//  Copyright 2014-2015 Manuel A. Fernandez Montecelo
+//  Copyright 2014-2016 Manuel A. Fernandez Montecelo
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -15,8 +15,8 @@
 //
 //  You should have received a copy of the GNU General Public License
 //  along with this program; see the file COPYING.  If not, write to
-//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-//  Boston, MA 02111-1307, USA.
+//  the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
+//  Boston, MA 02110-1301, USA.
 
 #include "aptitude.h"
 
@@ -98,42 +98,29 @@ cw::column_disposition pkg_ver_columnizer::setup_column(const pkgCache::VerItera
 	return cw::column_disposition(_("<N/A>"), 0);
 
       break;
-
     case pin_priority:
       {
-	if(apt_cache_file && !ver.end())
+	// sibling implementation of the one in pkg_columnizer.cc
+
+	// empty by default
+	string pin_priority_str;
+
+	pkgPolicy* policy = dynamic_cast<pkgPolicy *>(&(*apt_cache_file)->GetPolicy());
+
+	if (apt_cache_file && policy && !ver.end())
 	  {
-	    char buf[256];
+	    signed short priority = policy->GetPriority(ver);
 
-	    pkgPolicy *policy=dynamic_cast<pkgPolicy *>(&(*apt_cache_file)->GetPolicy());
-
-	    if(!policy)
-	      return cw::column_disposition("", 0);
-
-	    pkgCache::VerFileIterator vf=ver.FileList();
-
-	    if(vf.end())
-	      return cw::column_disposition("", 0);
-
-	    signed short priority=policy->GetPriority(vf.File());
-
-	    ++vf;
-
-	    // Find the highest priority for this version.
-	    while(!vf.end())
+	    if (priority != 0)
 	      {
-		priority=max(priority, policy->GetPriority(vf.File()));
-		++vf;
+		pin_priority_str = std::to_string(priority);
 	      }
-
-	    snprintf(buf, 256, "%d", priority);
-
-	    return cw::column_disposition(buf, 0);
 	  }
-	else
-	  return cw::column_disposition("", 0);
-      }
 
+	// return whatever was gathered
+	return cw::column_disposition(pin_priority_str, 0);
+      }
+      break;
     case sizechange:
       if(ver.end())
 	return cw::column_disposition("", 0);
@@ -339,41 +326,6 @@ cw::column_disposition pkg_ver_columnizer::setup_column(const pkgCache::VerItera
 	return cw::column_disposition("", 0);
 
       return cw::column_disposition(apt_package_records->Lookup(ver.FileList()).Maintainer(), 0);
-    case priority:
-      if(ver.end())
-	return cw::column_disposition("", 0);
-
-      if(const_cast<pkgCache::VerIterator &>(ver).PriorityType() &&
-	 const_cast<pkgCache::VerIterator &>(ver).PriorityType()[0])
-	return cw::column_disposition(const_cast<pkgCache::VerIterator &>(ver).PriorityType(), 0);
-      else
-	return cw::column_disposition(_("Unknown"), 0);
-    case shortpriority:
-      if(ver.end())
-	return cw::column_disposition("", 0);
-
-      switch(ver->Priority)
-	{
-	case pkgCache::State::Important:
-	  // TRANSLATORS: Imp = Important
-	  return cw::column_disposition(_("Imp"), 0);
-	case pkgCache::State::Required:
-	  // TRANSLATORS: Req = Required
-	  return cw::column_disposition(_("Req"), 0);
-	case pkgCache::State::Standard:
-	  // TRANSLATORS: Std = Standard
-	  return cw::column_disposition(_("Std"), 0);
-	case pkgCache::State::Optional:
-	  // TRANSLATORS: Opt = Optional
-	  return cw::column_disposition(_("Opt"), 0);
-	case pkgCache::State::Extra:
-	  // TRANSLATORS: Xtr = Extra
-	  return cw::column_disposition(_("Xtr"), 0);
-	default:
-	  return cw::column_disposition(_("ERR"), 0);
-	}
-
-      break;
     case section:
       if(ver.end())
 	return cw::column_disposition("", 0);
@@ -422,6 +374,13 @@ cw::column_disposition pkg_ver_columnizer::setup_column(const pkgCache::VerItera
 	else
 	  return cw::column_disposition("U", 0);
       }
+
+    case priority:
+    case shortpriority:
+    case source:
+    case architecture:
+    case origin:
+      // fall through
     default:
       return pkg_columnizer::setup_column(ver.ParentPkg(),
 					  ver,
@@ -648,7 +607,15 @@ void pkg_ver_item::hold(undo_group *undo)
 
 void pkg_ver_item::keep(undo_group *undo)
 {
-  // Do nothing for now.
+  // copied from pkg_item::keep()
+
+  auto package = version.ParentPkg();
+
+  // Keep, don't hold, the package.
+  (*apt_cache_file)->mark_keep(package,
+			       is_auto_installed(package),
+			       false,
+			       undo);
 }
 
 void pkg_ver_item::purge(undo_group *undo)
@@ -725,39 +692,6 @@ bool pkg_ver_item::dispatch_key(const cw::config::key &k, cw::tree *owner)
   else if(bindings->key_matches(k, "Changelog"))
     {
       view_changelog(version);
-      return true;
-    }
-  else if(bindings->key_matches(k, "BugReport"))
-    {
-      // Try to report a bug on the package.  (ew quoting ew)
-      string cmd=string("reportbug '")+version.ParentPkg().Name()+"' -V '"+version.VerStr()+"'";
-
-
-      struct sigaction oldact;
-      struct sigaction act;
-
-      memset(&act,0,sizeof(act));
-      act.sa_handler = SIG_DFL;
-      sigemptyset(&act.sa_mask);
-
-      sigaction(SIGCONT, &act, &oldact);
-
-      cw::toplevel::suspend();
-
-      apt_cache_file->ReleaseLock();
-
-      printf(_("Reporting a bug in %s:\n"), version.ParentPkg().Name());
-
-      if(system(cmd.c_str()) != 0) { /* ignore */ }
-
-      sigaction(SIGCONT, &oldact, NULL);
-
-      cw::toplevel::resume();
-
-      progress_ref p = gen_progress_bar();
-      apt_reload_cache(p->get_progress().unsafe_get_ref(), true);
-      p->destroy();
-
       return true;
     }
   else
@@ -808,4 +742,3 @@ bool pkg_ver_item::package_information()
   show_information();
   return true;
 }
-
